from __future__ import annotations
import types
import numpy as np
from functools import wraps
import warnings

from discretize.utils import as_array_n_by_dim  # noqa: F401

# scooby is a soft dependency for SimPEG
try:
    from scooby import Report as ScoobyReport
except ImportError:

    class ScoobyReport:
        def __init__(self, additional, core, optional, ncol, text_width, sort):
            print(
                "\n  *ERROR*: `SimPEG.Report` requires `scooby`."
                "\n           Install it via `pip install scooby` or"
                "\n           `conda install -c conda-forge scooby`.\n"
            )


def requires(var):
    """Wrap a function to require a specfic attribute.


    Use the following syntax to wrap a funciton::

        @requires('prob')
        def dpred(self):
            pass

    This wrapper will ensure that a problem has been bound to the data.
    If a problem is not bound an Exception will be raised, and an nice error message printed.

    Parameters
    ----------
    var :
        Input variable

    Returns
    -------
    wrapper
        The wrapper
    """

    def requiresVar(f):
        if var == "prob":
            extra = """

        .. note::

            To use survey.{0!s}(), SimPEG requires that a problem be bound to the survey.
            If a problem has not been bound, an Exception will be raised.
            To bind a problem to the Data object::

                survey.pair(myProblem)

            """.format(
                f.__name__
            )
        else:
            extra = """
                To use *{0!s}* method, SimPEG requires that the {1!s} be specified.
            """.format(
                f.__name__, var
            )

        @wraps(f)
        def requiresVarWrapper(self, *args, **kwargs):
            if getattr(self, var, None) is None:
                raise Exception(extra)
            return f(self, *args, **kwargs)

        doc = requiresVarWrapper.__doc__
        requiresVarWrapper.__doc__ = ("" if doc is None else doc) + extra

        return requiresVarWrapper

    return requiresVar


@requires("memory_profiler")
def mem_profile_class(input_class, *args):
    """Creates a new class from the target class with memory profiled methods.

    Using :meth:`memory_profiler.profile`, this function creates a wrapper class
    from the input class with the specified methods memory profiled.

    Parameters
    ----------
    input_class : class
        Input class being used to create the wrapper
    *args : str
        Method names that will be wrapped with a memory profiler. These names must
        correspond to methods of the input class.

    Returns
    -------
    class :
        Memory profiled class.

    Examples
    --------

    >>> FooMem = mem_profile_class(Foo,'my_func')
    >>> fooi = FooMem()
    >>> for i in range(5):
    >>>     fooi.my_func()

    Then run it from the command line

    ``python -m memory_profiler exampleMemWrapper.py``
    """
    from memory_profiler import profile

    attrs = {}
    for f in args:
        if hasattr(input_class, f):
            attrs[f] = profile(getattr(input_class, f))
        else:
            print("{0!s} not found in {1!s} Class".format(f, input_class.__name__))

    return type(input_class.__name__ + "MemProfileWrap", (input_class,), attrs)


def hook(obj, method, name=None, overwrite=False, silent=False):
    """Dynamically bind a class's method to an instance of a different class.

    Parameters
    ----------
    obj : class
        Instance of a class that will be binded to a new method
    method : method
        The method that will be binded to *obj*. The syntax is *ClassName.method*
    name : str, optional
        Provide a different name for the method being binded to *obj*. If ``None``,
        the original method name is used.
    overwrite : bool, default: ``False``
        If ``True``, the hook will overwrite a preexisting method of *obj* if it has
        the same name as the *name* input argument. If ``False``, preexisting methods
        are not overwritten.
    silent : bool, default: ``False``
        Print whether a previous hook was overwritten
    """
    if name is None:
        name = method.__name__
        if name == "<lambda>":
            raise Exception("Must provide name to hook lambda functions.")
    if not hasattr(obj, name) or overwrite:
        setattr(obj, name, types.MethodType(method, obj))
        if getattr(obj, "debug", False):
            print("Method " + name + " was added to class.")
    elif not silent or getattr(obj, "debug", False):
        print("Method " + name + " was not overwritten.")


def set_kwargs(obj, ignore=None, **kwargs):
    """
    Set key word arguments for an object or throw an error if any do not exist.

    Parameters
    ----------
    obj : class
        Instance of a class
    ignore : list, optional
        ``list`` of ``str`` denoting kwargs that are ignored (not being set)
    **kwargs
        Keyword arguments to set on the object.
    """
    if ignore is None:
        ignore = []
    for attr in kwargs:
        if attr in ignore:
            continue
        if hasattr(obj, attr):
            setattr(obj, attr, kwargs[attr])
        else:
            raise Exception("{0!s} attr is not recognized".format(attr))

    # hook(obj, hook, silent=True)
    # hook(obj, setKwargs, silent=True)


def print_done(obj, printers, name="Done", pad=""):
    """Print completion of an operation.

    (**DOCSTRING INCOMPLETE**)

    Parameters
    ----------
    obj : object
        An object
    printers : list of dict
        Has keys "width" and "title"
    name : str, default: "Done"
        A string for the process being completed
    pad : str, default: ""
        Trailing string

    """
    titles = ""
    widths = 0
    for printer in printers:
        titles += ("{{:^{0:d}}}".format(printer["width"])).format(printer["title"]) + ""
        widths += printer["width"]
    print(pad + "{0} {1} {0}".format("=" * ((widths - 1 - len(name)) // 2), name))
    # print(pad + "%s" % '-'*widths)


def print_titles(obj, printers, name="Print Titles", pad=""):
    """Print titles.

    (**DOCSTRING INCOMPLETE**)

    Parameters
    ----------
    obj : object
        An object
    printers : list of dict
        Has keys "width" and "title"
    name : str, default: "Print Titles"
        A string for the process being completed
    pad : str, default: ""
        Trailing string
    """
    titles = ""
    widths = 0
    for printer in printers:
        titles += ("{{:^{0:d}}}".format(printer["width"])).format(printer["title"]) + ""
        widths += printer["width"]
    print(pad + "{0} {1} {0}".format("=" * ((widths - 1 - len(name)) // 2), name))
    print(pad + titles)
    print(pad + "%s" % "-" * widths)


def print_line(obj, printers, pad=""):
    """Print line.

    (**DOCSTRING INCOMPLETE**)

    Parameters
    ----------
    obj : object
        An object
    printers : list of dict
        Dictonaries each have keys "width" and "title"
    pad : str, default: ""
        Trailing string
    """
    values = ""
    for printer in printers:
        values += ("{{:^{0:d}}}".format(printer["width"])).format(
            printer["format"] % printer["value"](obj)
        )
    print(pad + values)


def check_stoppers(obj, stoppers):
    """Check stopping rules.

    (**DOCSTRING INCOMPLETE**)

    Parameters
    ----------
    obj : object
        Input object
    stoppers : list of dict
        List of stoppers

    Returns
    -------
    bool
        Whether stopping criteria was encountered
    """
    optimal = []
    critical = []
    for stopper in stoppers:
        l = stopper["left"](obj)
        r = stopper["right"](obj)
        if stopper["stopType"] == "optimal":
            optimal.append(l <= r)
        if stopper["stopType"] == "critical":
            critical.append(l <= r)

    if obj.debug:
        print("checkStoppers.optimal: ", optimal)
    if obj.debug:
        print("checkStoppers.critical: ", critical)

    return (len(optimal) > 0 and all(optimal)) | (len(critical) > 0 and any(critical))


def print_stoppers(obj, stoppers, pad="", stop="STOP!", done="DONE!"):
    """Print stoppers.

    (**DOCSTRING INCOMPLETE**)

    Parameters
    ----------
    obj : object
        An object
    stoppers : list of dict
        Has keys "width" and "title"
    pad : str, default: ""
        Trailing string
    stop : str, default: "STOP!"
        String for statement when stopping criteria encountered
    done : str, default: "DONE!"
        String for statement when stopping criterian not encountered
    """
    print(pad + "{0!s}{1!s}{2!s}".format("-" * 25, stop, "-" * 25))
    for stopper in stoppers:
        l = stopper["left"](obj)
        r = stopper["right"](obj)
        print(pad + stopper["str"] % (l <= r, l, r))
    print(pad + "{0!s}{1!s}{2!s}".format("-" * 25, done, "-" * 25))


def call_hooks(match, mainFirst=False):
    """Wrap a function to an instance of a class.

    Use the following syntax::

        @callHooks('doEndIteration')
        def doEndIteration(self):
            pass

    This will call everything named *_doEndIteration* at the beginning of the function call.
    By default the main method (doEndIteration) is run after all of the sub methods (_doEndIteration*).
    This can be reversed by adding the mainFirst=True kwarg.

    Parameters
    ----------
    match : str
        Name of the function being wrapped to class instance
    mainFirst : bool, default: ``False``
        Main first

    Returns
    -------
    wrapper
        The wrapper
    """

    def callHooksWrap(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if not mainFirst:
                for method in [
                    posible for posible in dir(self) if ("_" + match) in posible
                ]:
                    if getattr(self, "debug", False):
                        print((match + " is calling self." + method))
                    getattr(self, method)(*args, **kwargs)

                return f(self, *args, **kwargs)
            else:
                out = f(self, *args, **kwargs)

                for method in [
                    posible for posible in dir(self) if ("_" + match) in posible
                ]:
                    if getattr(self, "debug", False):
                        print((match + " is calling self." + method))
                    getattr(self, method)(*args, **kwargs)

                return out

        extra = """
            If you have things that also need to run in the method {0!s}, you can create a method::

                def _{1!s}*(self, ... ):
                    pass

            Where the * can be any string. If present, _{2!s}* will be called at the start of the default {3!s} call.
            You may also completely overwrite this function.
        """.format(
            match, match, match, match
        )
        doc = wrapper.__doc__
        wrapper.__doc__ = ("" if doc is None else doc) + extra
        return wrapper

    return callHooksWrap


def dependent_property(name, value, children, doc):
    """Dependent property.

    Creates a property that mirrors a second property name.

    Parameters
    ----------
    name : str
        Property name
    value : scalar
        A scalar value
    children : class instances
        Child classes
    doc : str
        Property documentation
    """

    def fget(self):
        return getattr(self, name, value)

    def fset(self, val):
        if (np.isscalar(val) and getattr(self, name, value) == val) or val is getattr(
            self, name, value
        ):
            return  # it is the same!
        for child in children:
            if hasattr(self, child):
                delattr(self, child)
        setattr(self, name, val)

    return property(fget=fget, fset=fset, doc=doc)


class Report(ScoobyReport):
    """Print date, time, and version information.

    Use scooby to print date, time, and package version information in any
    environment (Jupyter notebook, IPython console, Python console, QT
    console), either as html-table (notebook) or as plain text (anywhere).

    Always shown are the OS, number of CPU(s), ``numpy``, ``scipy``,
    ``SimPEG``, ``cython``, ``properties``, ``vectormath``, ``discretize``,
    ``pymatsolver``, ``sys.version``, and time/date.

    Additionally shown are, if they can be imported, ``IPython``,
    ``matplotlib``, and ``ipywidgets``. It also shows MKL information, if
    available.

    All modules provided in ``add_pckg`` are also shown.


    Parameters
    ----------
    add_pckg : packages, optional
        Package or list of packages to add to output information (must be
        imported beforehand).
    ncol : int, optional
        Number of package-columns in html table (no effect in text-version);
        Defaults to 3.
    text_width : int, optional
        The text width for non-HTML display modes
    sort : bool, optional
        Sort the packages when the report is shown

    Examples
    --------

    >>> import pytest
    >>> import dateutil
    >>> from SimPEG import Report
    >>> Report()                            # Default values
    >>> Report(pytest)                      # Provide additional package
    >>> Report([pytest, dateutil], ncol=5)  # Define nr of columns

    """

    def __init__(self, add_pckg=None, ncol=3, text_width=80, sort=False):
        """Initiate a scooby.Report instance."""

        # Mandatory packages.
        core = [
            "SimPEG",
            "discretize",
            "pymatsolver",
            "numpy",
            "scipy",
            "sklearn",
            "matplotlib",
            "empymod",
            "geoana",
            "pandas",
        ]

        # Optional packages.
        optional = [
            "cython",
            "pydiso",
            "numba",
            "dask",
            "sympy",
            "IPython",
            "ipywidgets",
            "plotly",
            "vtk",
            "utm",
            "memory_profiler",
        ]

        super().__init__(
            additional=add_pckg,
            core=core,
            optional=optional,
            ncol=ncol,
            text_width=text_width,
            sort=sort,
        )


##############################################################
#               DEPRECATION FUNCTIONS
##############################################################


def deprecate_class(
    removal_version=None, new_location=None, future_warn=False, error=False
):
    """Utility function to deprecate a class

    Parameters
    ----------
    removal_version : str
        A string denoting the SimPEG version in which the class will be removed
    new_location : str
        Name for the class replacing the deprecated class
    future_warn : bool, default: ``False``
        If ``True``, throw comprehensive warning the class will be deprecated
    error : bool, default: ``False``
        Throw error if deprecated class no longer implemented

    Returns
    -------
    class
        The new class
    """

    def decorator(cls):
        my_name = cls.__name__
        parent_name = cls.__bases__[0].__name__
        message = f"{my_name} has been deprecated, please use {parent_name}."
        if error:
            message = f"{my_name} has been removed, please use {parent_name}."
        elif removal_version is not None:
            message += f" It will be removed in version {removal_version} of SimPEG."
        else:
            message += " It will be removed in a future version of SimPEG."

        # stash the original initialization of the class
        cls._old__init__ = cls.__init__

        def __init__(self, *args, **kwargs):
            if future_warn:
                warnings.warn(message, FutureWarning)
            elif error:
                raise NotImplementedError(message)
            else:
                warnings.warn(message, DeprecationWarning)
            self._old__init__(*args, **kwargs)

        cls.__init__ = __init__
        if new_location is not None:
            parent_name = f"{new_location}.{parent_name}"
        cls.__doc__ = f""" This class has been deprecated, see `{parent_name}` for documentation"""
        return cls

    return decorator


def deprecate_module(
    old_name, new_name, removal_version=None, future_warn=False, error=False
):
    """Deprecate module

    Parameters
    ----------
    old_name : str
        Original name for the now deprecated module
    new_name : str
        New name for the module
    removal_version : str, optional
        SimPEG version in which the module will be removed from the code base
    future_warn : bool, default: ``False``
        If ``True``, throw comprehensive warning the module will be deprecated
    error : bool, default: ``False``
        Throw error if deprecated module no longer implemented
    """
    message = f"The {old_name} module has been deprecated, please use {new_name}."
    if error:
        message = f"{old_name} has been removed, please use {new_name}."
    elif removal_version is not None:
        message += f" It will be removed in version {removal_version} of SimPEG"
    else:
        message += " It will be removed in a future version of SimPEG."
    message += " Please update your code accordingly."
    if future_warn:
        warnings.warn(message, FutureWarning)
    elif error:
        raise NotImplementedError(message)
    else:
        warnings.warn(message, DeprecationWarning)


def deprecate_property(
    prop, old_name, new_name=None, removal_version=None, future_warn=False, error=False
):
    """Deprecate property

    Parameters
    ----------
    prop : property
        Current property
    old_name : str
        Original name for the now deprecated property
    new_name : str, optional
        New name for the property. If ``None``, the property name is take from the
        *prop* input argument.
    removal_version : str, optional
        SimPEG version in which the property will be removed from the code base
    future_warn : bool, default: ``False``
        If ``True``, throw comprehensive warning the property will be deprecated
    error : bool, default: ``False``
        Throw error if deprecated property no longer implemented

    Returns
    -------
    property
        The new property
    """

    if isinstance(prop, property):
        if new_name is None:
            new_name = prop.fget.__qualname__
        cls_name = new_name.split(".")[0]
        old_name = f"{cls_name}.{old_name}"

    message = f"{old_name} has been deprecated, please use {new_name}."
    if error:
        message = f"{old_name} has been removed, please use {new_name}."
    elif removal_version is not None:
        message += f" It will be removed in version {removal_version} of SimPEG."
    else:
        message += " It will be removed in a future version of SimPEG."

    def get_dep(self):
        if future_warn:
            warnings.warn(message, FutureWarning)
        elif error:
            raise NotImplementedError(message)
        else:
            warnings.warn(message, DeprecationWarning)
        return prop.fget(self)

    def set_dep(self, other):
        if future_warn:
            warnings.warn(message, FutureWarning)
        elif error:
            raise NotImplementedError(message)
        else:
            warnings.warn(message, DeprecationWarning)
        prop.fset(self, other)

    doc = f"`{old_name}` has been deprecated. See `{new_name}` for documentation"

    return property(get_dep, set_dep, prop.fdel, doc)


def deprecate_method(
    method, old_name, removal_version=None, future_warn=False, error=False
):
    """Deprecate method

    Parameters
    ----------
    method : method
        Current method
    old_name : str
        Original name for the now deprecated method
    removal_version : str, optional
        SimPEG version in which the method will be removed from the code base
    future_warn : bool, default: ``False``
        If ``True``, throw comprehensive warning the method will be deprecated
    error : bool, default: ``False``
        Throw error if deprecated method no longer implemented

    Returns
    -------
    method
        The new method
    """
    new_name = method.__qualname__
    split_name = new_name.split(".")
    if len(split_name) > 1:
        old_name = f"{split_name[0]}.{old_name}"

    message = f"{old_name} has been deprecated, please use {new_name}."
    if error:
        message = f"{old_name} has been removed, please use {new_name}."
    elif removal_version is not None:
        message += f" It will be removed in version {removal_version} of SimPEG."
    else:
        message += " It will be removed in a future version of SimPEG."

    def new_method(*args, **kwargs):
        if future_warn:
            warnings.warn(message, FutureWarning)
        elif error:
            raise NotImplementedError(message)
        else:
            warnings.warn(message, DeprecationWarning)
        return method(*args, **kwargs)

    doc = f"`{old_name}` has been deprecated. See `{new_name}` for documentation"
    new_method.__doc__ = doc
    return new_method


def deprecate_function(
    new_function,
    old_name,
    removal_version=None,
    new_location=None,
    future_warn=False,
    error=False,
):
    """Deprecate function

    Parameters
    ----------
    new_function : function
        Current function
    old_name : str
        Original name for the now deprecated function
    removal_version : str, optional
        SimPEG version in which the method will be removed from the code base
    future_warn : bool, default: ``False``
        If ``True``, throw comprehensive warning the method will be deprecated
    error : bool, default: ``False``
        Throw error if deprecated method no longer implemented

    Returns
    -------
    function
        The new function
    """
    new_name = new_function.__name__
    if new_location is not None:
        new_name = f"{new_location}.{new_name}"

    message = f"{old_name} has been deprecated, please use {new_name}."
    if error:
        message = f"{old_name} has been removed, please use {new_name}."
    elif removal_version is not None:
        message += f" It will be removed in version {removal_version} of SimPEG."
    else:
        message += " It will be removed in a future version of SimPEG."

    def dep_function(*args, **kwargs):
        if future_warn:
            warnings.warn(message, FutureWarning)
        elif error:
            raise NotImplementedError(message)
        else:
            warnings.warn(message, DeprecationWarning)
        return new_function(*args, **kwargs)

    doc = f"""
    `{old_name}` has been deprecated. See `{new_name}` for documentation

    See Also
    --------
    {new_name}
    """
    dep_function.__doc__ = doc
    return dep_function


###############################################################
#                    PROPERTY VALIDATORS
###############################################################


def validate_string(property_name, var, string_list=None, case_sensitive=False):
    """Validate a string property

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : str
        The input variable
    string_list : list or tuple of str, optional
        Provide a list of acceptable strings, if an individual item is also a list,
        the extra parameters are interpreted as aliases for the first item, which is then returned.
    case_sensitive : bool, default: ``False``
        If ``True`` and *string_list* is not ``None``, the string
        comparison is case-sensitive

    Returns
    -------
    str
        Returns the input argument *var* once validated
    """
    if isinstance(var, str):
        if string_list is None:
            return var
        if not case_sensitive:
            test_var = var.casefold()

            # also fold the string_list for comparison
            def fold_input(input_variable):
                if isinstance(input_variable, (list, tuple)):
                    return [fold_input(x) for x in input_variable]
                return input_variable.casefold()

            test_string_list = fold_input(string_list)
        else:
            test_var = var
            test_string_list = string_list

        for test, item in zip(test_string_list, string_list):
            if isinstance(test, (list, tuple)):
                if test_var in test:
                    return item[0]
            if test_var == test:
                return item
        raise ValueError(f"{property_name!r} must be in {string_list!r}. Got {var!r}")
    else:
        raise TypeError(f"{property_name!r} must be a str. Got {type(var)}")


def validate_integer(property_name, var, min_val=-np.inf, max_val=np.inf):
    """Validate integer property

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : int or float
        The input variable
    min_val : int or float, optional
        Minimum value
    max_val : int or float, optional
        Maximum value

    Returns
    -------
    int
        Returns the input variable as a int once validated

    Notes
    -----
    The variable is casted to an `int`, therefore if a float is passed, the decimal
    value is discarded and a valid integer is returned.
    """
    try:
        var = int(var)
    except (ValueError, TypeError) as err:
        raise TypeError(f"{property_name!r} must be a number, got {type(var)}") from err

    if (var < min_val) | (var > max_val):
        raise ValueError(
            f"{property_name!r} must be a value between {min_val} and {max_val}"
        )
    else:
        return var


def validate_float(
    property_name,
    var,
    min_val=-np.inf,
    max_val=np.inf,
    inclusive_min=True,
    inclusive_max=True,
):
    """Validate float property

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : number
        The input variable which will be cast to a float
    min_val, max_val : int or float, optional
        Minimum/Maximum value
    inclusive_min, inclusive_max : bool, optional
        Whether the minimum and maximum values are inclusive.

    Returns
    -------
    float
        Returns the input variable as a float once validated

    Notes
    -----
    The input is first attempted ot be cast to a float.
    """
    try:
        var = float(var)
    except (ValueError, TypeError) as err:
        raise TypeError(
            f"{property_name!r} must be int or float, got {type(var)}"
        ) from err

    value_range_string = f"{min_val}, {max_val}"
    if inclusive_min:
        value_range_string = "[" + value_range_string
    else:
        value_range_string = "(" + value_range_string
    if inclusive_max:
        value_range_string = value_range_string + "]"
    else:
        value_range_string = value_range_string + ")"

    if (
        (inclusive_min and var < min_val)
        or (not inclusive_min and var <= min_val)
        or (inclusive_max and var > max_val)
        or (not inclusive_max and var >= max_val)
    ):
        raise ValueError(
            f"{property_name!r} must be a value in the range " + value_range_string
        )
    else:
        return var


def validate_list_of_types(property_name, var, class_type, ensure_unique=False):
    """Validate list of instances of a certain class

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : object or a list of object
        A list of objects
    class_type : class or tuple of class types
        Class type(s) that are allowed in the list
    ensure_unique : bool, optional
        Checks if all items in the var are unique items.

    Returns
    -------
    list
        Returns the list once validated
    """
    if isinstance(var, list):
        pass
    elif isinstance(var, class_type):
        var = [var]
    else:
        raise TypeError(f"{property_name!r} must be a list of {class_type}")

    is_true = [isinstance(x, class_type) for x in var]
    if np.all(is_true):
        if ensure_unique and len(set(var)) != len(var):
            raise ValueError(
                f"The {property_name!r} list must be unique. Cannot re-use items"
            )
        return var
    else:
        raise TypeError(f"{property_name!r} must be a list of {class_type}")


def validate_location_property(property_name, var, dim=None):
    """Validate a location

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : array_like
        The input variable
    dim : int, optional
        The dimension; i.e. 1, 2 or 3

    Returns
    -------
    numpy.ndarray
        Returns the location once validated
    """
    try:
        var = np.atleast_1d(var).astype(float).squeeze()
    except (TypeError, ValueError) as err:
        raise TypeError(
            f"{property_name!r} must be 1D array_like, got {type(var)}"
        ) from err

    if len(var.shape) > 1:
        raise ValueError(
            f"{property_name!r} must be 1D array_like, got {len(var.shape)}D"
        )

    if dim is None:
        return var
    else:
        if len(var) == dim:
            return var
        else:
            raise ValueError(
                f"{property_name!r} must be array_like with shape ({dim}, ), got ({len(var)},)"
            )


def validate_ndarray_with_shape(property_name, var, shape=None, dtype=float):
    """Validate numerical array property

    Parameters
    ----------
    property_name : str
        The name of the property being set
    var : array_like
        The input array
    shape : tuple of int, or list of tuple of int, default: None
        The shape of the array; e.g. (3, ), (3, 3), ('*', 2).
        The '*' indicates that an arbitrary number of elements is allowed
        along a particular dimension. If list then multiple shapes are accepted.
        By default, shape is a tuple of length ndim of '*'.
    dtype : class, or tuple of class, optional
        The data type for the array. I.e. float, int, complex, bool, etc.
        Validated from left to right if a tuple.

    Returns
    -------
    numpy.ndarray of dtype
        Returns the array in the specified data type once validated
    """
    if not isinstance(dtype, tuple):
        dtypes = (dtype,)
    else:
        dtypes = dtype
    for dtype in dtypes:
        try:
            var = np.asarray(var, dtype=dtype)
            bad_type = False
            break
        except (TypeError, ValueError) as err:
            bad_type = True
            raised_err = err

    if bad_type:
        raise TypeError(
            f"{property_name!r} must be array_like with data type of {dtype}, got {type(var)}"
        ) from raised_err

    if shape is None:
        return var

    if isinstance(shape, tuple):
        shapes = [
            shape,
        ]
    else:
        shapes = shape
    for shp in shapes:
        if len(shp) > 3:
            raise NotImplementedError("Only implemented for 1D, 2D and 3D arrays!!!")

        shape_error = False
        dim_error = var.ndim > len(shp)
        if not dim_error:
            if len(shp) == 1:
                var_array = np.atleast_1d(var)
            elif len(shp) == 2:
                var_array = np.atleast_2d(var)
            elif len(shp) == 3:
                var_array = np.atleast_3d(var)

            for ii, value in enumerate(np.shape(var_array)):
                if (shp[ii] != "*") & (shp[ii] != value):
                    shape_error = True
            if not shape_error:
                return var_array
    if dim_error:
        raise ValueError(
            "The dimensions of the input argument 'var' are greater than is specified by 'shape'"
        )
    if shape_error:
        if isinstance(shape, list):
            raise ValueError(
                f"{property_name!r} must be one of {shape}, got {np.shape(var)}"
            )
        else:
            raise ValueError(f"{property_name!r} must be {shape}, got {np.shape(var)}")


def validate_type(property_name, obj, obj_type, cast=True, strict=False):
    """Validate the type or an item

    Parameters
    ----------
    property_name : str
        The name of the property being set
    obj : object
        The object to test
    obj_type : class
        The intended type of the object
    cast : bool, optional
        Whether to attempt to cast the item to the class
    strict : bool, optional
        Whether to test if the object is exactly the type, (or a subclass if false)

    Returns
    -------
    obj_type
        Returns the object in the specified type when validated
    """
    if cast:
        try:
            obj = obj_type(obj)
        except Exception as err:
            raise TypeError(
                f"{type(obj).__name__} cannot be converted to type {obj_type.__name__} "
                f"required for {property_name}."
            ) from err
    if strict and type(obj) != obj_type:
        raise TypeError(
            f"Object must be exactly a {obj_type.__name__} for {property_name}"
        )
    if not isinstance(obj, obj_type):
        raise TypeError(
            f"Object must be an instance of {obj_type.__name__} for {property_name}"
        )
    return obj


def validate_callable(property_name, obj):
    """
    Validate if an object is callable

    Parameters
    ----------
    property_name : str
        The name of the property being set
    obj : object
        The object to test

    Returns
    -------
    callable
    """
    if not callable(obj):
        raise TypeError(f"{property_name} must be callable and {type(obj)} is not.")
    return obj


def validate_direction(property_name, obj, dim=3):
    """
    Validate if an object represents a direction.

    Parameters
    ----------
    property_name : str
        The name of the property being set
    obj : {"x", "y", "z"} or (dim) numpy.ndarray
        The object to test
    dim : int, optional
        The dimension of the vector

    Returns
    -------
    (dim) numpy.ndarray
    """

    if isinstance(obj, str):
        obj = validate_string(property_name, obj, string_list=("x", "y", "z")[:dim])
        if obj == "x":
            obj = np.r_[1.0, 0.0, 0.0][:dim]
        elif obj == "y":
            obj = np.r_[0.0, 1.0, 0.0][:dim]
        elif obj == "z":
            obj = np.r_[0.0, 0.0, 1.0][:dim]

    obj = validate_ndarray_with_shape(property_name, obj, shape=(dim,), dtype=float)

    # Normalize the orientation
    # do this to make a copy of the input
    obj = obj / np.linalg.norm(obj)
    return obj


def validate_active_indices(property_name, index_arr, n_cells):
    """Validates an array intended to indicate a list of active indices.

    This accepts either an array of booleans, or an array of active integers.

    Parameters
    ----------
    property_name : str
        The name of the property
    index_arr : array_like
        The active index array to validate
    n_cells : int
        The length of the expected array

    Returns
    -------
    (n_cells) numpy.ndarray of bool

    """
    index_arr = np.asarray(index_arr)
    index_arr = np.atleast_1d(np.squeeze(index_arr))
    if index_arr.ndim != 1:
        raise ValueError(f"{property_name} must be a 1D array.")

    if not np.issubdtype(index_arr.dtype, bool):
        if not np.issubdtype(index_arr.dtype, np.integer):
            raise TypeError(
                f"{property_name} must be an array of integers (or boolean) "
                "indicating the active cells."
            )

        tmp = np.zeros((n_cells,), dtype=bool)
        if index_arr.max() >= n_cells:
            raise IndexError(
                f"maximum index {index_arr.max()} out of bounds for `{property_name}` with size {n_cells}"
            )
        tmp[index_arr] = True

        if np.sum(tmp) != len(index_arr):
            # This line should cause an error to be thrown if someone
            # accidentally passes a list of 0 & 1 integers instead of passing
            # it a list of booleans.
            raise ValueError(
                f"{property_name} was interpreted as a list of active indices and you "
                "attempted to set the same cell as active multiple times."
            )
        index_arr = tmp

    if index_arr.shape != (n_cells,):
        raise ValueError(f"Input 'active_cells' must have shape {(n_cells,)}")
    return index_arr


###############################################################
#                      DEPRECATIONS
###############################################################
memProfileWrapper = deprecate_function(
    mem_profile_class, "memProfileWrapper", removal_version="0.18.0"
)
setKwargs = deprecate_function(set_kwargs, "setKwargs", removal_version="0.18.0")
printTitles = deprecate_function(print_titles, "printTitles", removal_version="0.18.0")
printLine = deprecate_function(print_line, "printLine", removal_version="0.18.0")
printStoppers = deprecate_function(
    print_stoppers, "printStoppers", removal_version="0.18.0"
)
checkStoppers = deprecate_function(
    check_stoppers, "checkStoppers", removal_version="0.18.0"
)
printDone = deprecate_function(print_done, "printDone", removal_version="0.18.0")
callHooks = deprecate_function(call_hooks, "callHooks", removal_version="0.18.0")
dependentProperty = deprecate_function(
    dependent_property, "dependentProperty", removal_version="0.18.0"
)
asArray_N_x_Dim = deprecate_function(
    as_array_n_by_dim, "asArray_N_x_Dim", removal_version="0.19.0", future_warn=True
)



from ... import survey
from ...utils import validate_string


class Point(survey.BaseRx):
    """Point receiver class for gravity simulations

    The **Point** receiver defines the locations and the components of the gravity
    field that are simulated at each location. The length of the resulting data
    vector is *n_loc X n_comp*, and is organized by location then component.

    Parameters
    ----------
    locations: (n_loc, 3) numpy.ndarray
        Receiver locations
    components: str or list of str
        Data component(s) measured at each receiver location. Use a ``str`` for a
        single component or a ``list`` of ``str`` if multiple components are simulated
        at each location. Component choices are:

        - "gx"   --> x-component of the gravity field
        - "gy"   --> y-component of the gravity field
        - "gz"   --> z-component of the gravity field (DEFAULT)
        - "gxx"  --> x-derivative of the x-component
        - "gxy"  --> y-derivative of the x-component (and visa versa)
        - "gxz"  --> z-derivative of the x-component (and visa versa)
        - "gyy"  --> y-derivative of the y-component
        - "gyz"  --> z-derivative of the y-component (and visa versa)
        - "gzz"  --> z-derivative of the z-component
        - "guv"  --> UV component
    """

    def __init__(self, locations, components="gz", **kwargs):
        super(Point, self).__init__(locations, **kwargs)

        if isinstance(components, str):
            components = [components]

        for component in components:
            validate_string(
                "component",
                component,
                [
                    "gx",
                    "gy",
                    "gz",
                    "gxx",
                    "gxy",
                    "gxz",
                    "gyy",
                    "gyz",
                    "gzz",
                    "guv",
                ],
            )
        self.components = components

    @property
    def nD(self):
        """Number of data

        Returns
        -------
        int
            The number of data
        """
        return self.locations.shape[0] * len(self.components)



"""
=================================================================
Base Classes and Functions (:mod:`SimPEG.potential_fields`)
=================================================================
.. currentmodule:: SimPEG.potential_fields

About ``potential_fields``

Base classes and utility functions
==================================

.. autosummary::
  :toctree: generated/

  base.BasePFSimulation
  base.BaseEquivalentSourceLayerSimulation
  base.progress
  get_dist_wgt


"""

from . import magnetics
from . import gravity

from .base import get_dist_wgt



from ...survey import BaseSrc


class SourceField(BaseSrc):
    """Source field for gravity integral formulation

    Parameters
    ----------
    receivers_list : list of SimPEG.potential_fields.receivers.Point
        List of magnetics receivers
    """

    def __init__(self, receiver_list=None, **kwargs):
        super(SourceField, self).__init__(receiver_list=receiver_list, **kwargs)



from .TDEM import hzAnalyticDipoleT, hzAnalyticCentLoopT
from .FDEM import hzAnalyticDipoleF
from .FDEMcasing import *
from .DC import *
from .FDEMDipolarfields import *
from .NSEM import MT_LayeredEarth



from __future__ import annotations

import numpy as np
from discretize.base import BaseMesh
import warnings
from typing import TYPE_CHECKING
from .. import maps
from ..objective_function import BaseObjectiveFunction, ComboObjectiveFunction
from .. import utils
from .regularization_mesh import RegularizationMesh

from SimPEG.utils.code_utils import deprecate_property, validate_ndarray_with_shape

import scipy.sparse as sp


class BaseRegularization(BaseObjectiveFunction):
    """
    Base class for regularization. Inherit this for building your own
    regularization. The base regularization assumes a weighted l2-norm style of
    regularization. However, if you wish to employ a different norm, the
    methods :meth:`__call__`, :meth:`deriv` and :meth:`deriv2` can be
    over-written

    :param discretize.base.BaseMesh mesh: SimPEG mesh
    :param active_cells: Array of bool defining the set of active cells.
    :param mapping: Model map
    :param reference_model: Array of model values used to constrain the inversion
    :param units: Model units identifier. Special case for 'radian'
    :param weights: Weight multipliers to customize the least-squares function.
    """

    def __init__(
        self,
        mesh: RegularizationMesh | BaseMesh,
        active_cells: np.ndarray | None = None,
        mapping: maps.IdentityMap | None = None,
        reference_model: np.ndarray | None = None,
        units: str | None = None,
        weights: dict | None = None,
        **kwargs,
    ):
        if isinstance(mesh, BaseMesh):
            mesh = RegularizationMesh(mesh)

        if not isinstance(mesh, RegularizationMesh):
            raise TypeError(
                f"'regularization_mesh' must be of type {RegularizationMesh} or {BaseMesh}. "
                f"Value of type {type(mesh)} provided."
            )

        self._model = None
        self._parent = None
        self._regularization_mesh = mesh
        self._weights = {}

        if active_cells is not None:
            self.active_cells = active_cells

        self.mapping = mapping

        super().__init__(**kwargs)

        self.reference_model = reference_model
        self.units = units

        if weights is not None:
            if not isinstance(weights, dict):
                weights = {"user_weights": weights}
            self.set_weights(**weights)

    @property
    def active_cells(self) -> np.ndarray:
        """A boolean array of active cells on the regularization

        Returns
        -------
        (n_cells, ) Array of bool

        Notes
        -----
        If this is set with an array of integers, it interprets it as an array
        listing the active cell indices.
        """
        return self.regularization_mesh.active_cells

    @active_cells.setter
    def active_cells(self, values: np.ndarray | None):
        self.regularization_mesh.active_cells = values

        if values is not None:
            volume_term = "volume" in self._weights
            self._weights = {}
            self._W = None
            if volume_term:
                self.set_weights(volume=self.regularization_mesh.vol)

    indActive = deprecate_property(
        active_cells,
        "indActive",
        "active_cells",
        "0.19.0",
        future_warn=True,
        error=False,
    )

    @property
    def model(self) -> np.ndarray:
        """Physical property model"""
        return self._model

    @model.setter
    def model(self, values: np.ndarray | float):
        if isinstance(values, float):
            values = np.ones(self._nC_residual) * values

        values = validate_ndarray_with_shape(
            "model", values, shape=(self._nC_residual,), dtype=float
        )

        self._model = values

    @property
    def mapping(self) -> maps.IdentityMap:
        """Mapping applied to the model values"""
        return self._mapping

    @mapping.setter
    def mapping(self, mapping: maps.IdentityMap):
        if mapping is None:
            mapping = maps.IdentityMap()
        if not isinstance(mapping, maps.IdentityMap):
            raise TypeError(
                f"'mapping' must be of type {maps.IdentityMap}. "
                f"Value of type {type(mapping)} provided."
            )
        self._mapping = mapping

    @property
    def parent(self):
        """
        The parent objective function
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        combo_class = ComboObjectiveFunction
        if not isinstance(parent, combo_class):
            raise TypeError(
                f"Invalid parent of type '{parent.__class__.__name__}'. "
                f"Parent must be a {combo_class.__name__}."
            )
        self._parent = parent

    @property
    def units(self) -> str | None:
        """Specify the model units. Special care given to 'radian' values"""
        return self._units

    @units.setter
    def units(self, units: str | None):
        if units is not None and not isinstance(units, str):
            raise TypeError(
                f"'units' must be None or type str. "
                f"Value of type {type(units)} provided."
            )
        self._units = units

    @property
    def _weights_shapes(self) -> tuple[int] | str:
        """Acceptable lengths for the weights

        Returns
        -------
        list of tuple
            Each tuple represents accetable shapes for the weights
        """
        if (
            getattr(self, "_regularization_mesh", None) is not None
            and self.regularization_mesh.nC != "*"
        ):
            return (self.regularization_mesh.nC,)

        if getattr(self, "_mapping", None) is not None and self.mapping.shape != "*":
            return (self.mapping.shape[0],)

        return ("*",)

    @property
    def reference_model(self) -> np.ndarray:
        """Reference physical property model"""
        return self._reference_model

    @reference_model.setter
    def reference_model(self, values: np.ndarray | float):
        if values is not None:
            if isinstance(values, float):
                values = np.ones(self._nC_residual) * values

            values = validate_ndarray_with_shape(
                "reference_model", values, shape=(self._nC_residual,), dtype=float
            )
        self._reference_model = values

    mref = deprecate_property(
        reference_model,
        "mref",
        "reference_model",
        "0.19.0",
        future_warn=True,
        error=False,
    )

    @property
    def regularization_mesh(self) -> RegularizationMesh:
        """Regularization mesh"""
        return self._regularization_mesh

    regmesh = deprecate_property(
        regularization_mesh,
        "regmesh",
        "regularization_mesh",
        "0.19.0",
        future_warn=True,
        error=False,
    )

    @property
    def cell_weights(self) -> np.ndarray:
        """Deprecated property for 'volume' and user defined weights."""
        warnings.warn(
            "cell_weights are deprecated please access weights using the `set_weights`,"
            " `get_weights`, and `remove_weights` functionality. This will be removed in 0.19.0",
            FutureWarning,
        )
        return np.prod(list(self._weights.values()), axis=0)

    @cell_weights.setter
    def cell_weights(self, value):
        warnings.warn(
            "cell_weights are deprecated please access weights using the `set_weights`,"
            " `get_weights`, and `remove_weights` functionality. This will be removed in 0.19.0",
            FutureWarning,
        )
        self.set_weights(cell_weights=value)

    def get_weights(self, key) -> np.ndarray:
        """Weights for a given key."""
        return self._weights[key]

    def set_weights(self, **weights):
        """Adds (or updates) the specified weights to the regularization

        Parameters:
        -----------
        **kwargs : key, numpy.ndarray
            Each keyword argument is added to the weights used by the regularization.
            They can be accessed with their keyword argument.

        Examples
        --------
        >>> import discretize
        >>> from SimPEG.regularization import Smallness
        >>> mesh = discretize.TensorMesh([2, 3, 2])
        >>> reg = Smallness(mesh)
        >>> reg.set_weights(my_weight=np.ones(mesh.n_cells))
        >>> reg.get_weights('my_weight')
        array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
        """
        for key, values in weights.items():
            values = validate_ndarray_with_shape(
                "weights", values, shape=self._weights_shapes, dtype=float
            )
            self._weights[key] = values
        self._W = None

    def remove_weights(self, key):
        """Removes the weights with a given key"""
        try:
            self._weights.pop(key)
        except KeyError as error:
            raise KeyError(f"{key} is not in the weights dictionary") from error
        self._W = None

    @property
    def W(self) -> np.ndarray:
        """
        Weighting matrix
        """
        if getattr(self, "_W", None) is None:
            weights = np.prod(list(self._weights.values()), axis=0)
            self._W = utils.sdiag(weights**0.5)
        return self._W

    @property
    def _nC_residual(self) -> int:
        """
        Shape of the residual
        """

        nC = getattr(self.regularization_mesh, "nC", None)
        mapping = getattr(self, "_mapping", None)

        if mapping is not None and mapping.shape[1] != "*":
            return self.mapping.shape[1]

        if nC != "*" and nC is not None:
            return self.regularization_mesh.nC

        return self._weights_shapes[0]

    def _delta_m(self, m) -> np.ndarray:
        if self.reference_model is None:
            return m
        return (
            m - self.reference_model
        )  # in case self.reference_model is Zero, returns type m

    @utils.timeIt
    def __call__(self, m):
        r"""
        We use a weighted 2-norm objective function

        .. math::

            r(m) = \frac{1}{2} \| \mathbf{W} \mathbf{f(m)} \|_2^2
        """
        r = self.W * self.f_m(m)
        return 0.5 * r.dot(r)

    def f_m(self, m) -> np.ndarray:
        raise AttributeError("Regularization class must have a 'f_m' implementation.")

    def f_m_deriv(self, m) -> sp.csr_matrix:
        raise AttributeError(
            "Regularization class must have a 'f_m_deriv' implementation."
        )

    @utils.timeIt
    def deriv(self, m) -> np.ndarray:
        r"""
        The regularization is:

        .. math::

            R(m) = \frac{1}{2}\mathbf{(m-m_\text{ref})^\top W^\top
                   W(m-m_\text{ref})}

        So the derivative is straight forward:

        .. math::

            R(m) = \mathbf{W^\top W (m-m_\text{ref})}

        """
        r = self.W * self.f_m(m)
        return self.f_m_deriv(m).T * (self.W.T * r)

    @utils.timeIt
    def deriv2(self, m, v=None) -> sp.csr_matrix:
        r"""
        Second derivative

        :param numpy.ndarray m: geophysical model
        :param numpy.ndarray v: vector to multiply
        :rtype: scipy.sparse.sp.csr_matrix
        :return: WtW, or if v is supplied WtW*v (numpy.ndarray)

        The regularization is:

        .. math::

            R(m) = \frac{1}{2}\mathbf{(m-m_\text{ref})^\top W^\top
            W(m-m_\text{ref})}

        So the second derivative is straight forward:

        .. math::

            R(m) = \mathbf{W^\top W}

        """
        f_m_deriv = self.f_m_deriv(m)
        if v is None:
            return f_m_deriv.T * ((self.W.T * self.W) * f_m_deriv)

        return f_m_deriv.T * (self.W.T * (self.W * (f_m_deriv * v)))


class Smallness(BaseRegularization):
    r"""
    Small regularization - L2 regularization on the difference between a
    model and a reference model.

    .. math::

        r(m) = \frac{1}{2}(\mathbf{m} - \mathbf{m_ref})^\top \mathbf{V}^T
            \mathbf{W}^T
        \mathbf{W} \mathbf{V} (\mathbf{m} - \mathbf{m_{ref}})

    where
    :math:`\mathbf{m}` is the model,
    :math:`\mathbf{m_{ref}}` is a reference model,
    :math:`\mathbf{V}` are square root of cell volumes and
    :math:`\mathbf{W}` is a weighting matrix (default Identity). If fixed or
        free weights are provided, then it is :code:`diag(np.sqrt(weights))`).


    **Optional Inputs**

    :param discretize.base.BaseMesh mesh: SimPEG mesh
    :param int shape: number of parameters
    :param IdentityMap mapping: regularization mapping, takes the model from
        model space to the space you want to regularize in
    :param numpy.ndarray reference_model: reference model
    :param numpy.ndarray active_cells: active cell indices for reducing the size
        of differential operators in the definition of a regularization mesh
    :param numpy.ndarray weights: cell weights

    """

    _multiplier_pair = "alpha_s"

    def __init__(self, mesh, **kwargs):
        super().__init__(mesh, **kwargs)
        self.set_weights(volume=self.regularization_mesh.vol)

    def f_m(self, m) -> np.ndarray:
        """
        Model residual
        """
        return self.mapping * self._delta_m(m)

    def f_m_deriv(self, m) -> sp.csr_matrix:
        """
        Derivative of the model residual
        """
        return self.mapping.deriv(self._delta_m(m))


class SmoothnessFirstOrder(BaseRegularization):
    """
    Smooth Regularization. This base class regularizes on the first
    spatial derivative, optionally normalized by the base cell size.

    **Optional Inputs**

    :param discretize.base.BaseMesh mesh: SimPEG mesh
    :param IdentityMap mapping: regularization mapping, takes the model from
        model space to the space you want to regularize in
    :param numpy.ndarray reference_model: reference model
    :param numpy.ndarray active_cells: active cell indices for reducing the
        size of differential operators in the definition of a regularization mesh
    :param numpy.ndarray weights: cell weights
    :param bool reference_model_in_smooth: include the reference model in the
        smoothness computation? (eg. look at Deriv of m (False) or Deriv of
        (m-reference_model) (True))
    :param numpy.ndarray weights: vector of cell weights (applied in all terms)
    """

    def __init__(
        self, mesh, orientation="x", reference_model_in_smooth=False, **kwargs
    ):
        self.reference_model_in_smooth = reference_model_in_smooth

        if orientation not in ["x", "y", "z"]:
            raise ValueError("Orientation must be 'x', 'y' or 'z'")

        if orientation == "y" and mesh.dim < 2:
            raise ValueError(
                "Mesh must have at least 2 dimensions to regularize along the "
                "y-direction."
            )
        elif orientation == "z" and mesh.dim < 3:
            raise ValueError(
                "Mesh must have at least 3 dimensions to regularize along the "
                "z-direction"
            )
        self._orientation = orientation
        self.__cell_distances = None
        super().__init__(mesh=mesh, **kwargs)
        self.set_weights(volume=self.regularization_mesh.vol)

    @property
    def _weights_shapes(self):
        """Acceptable lengths for the weights

        Returns
        -------
        tuple
            A tuple of each acceptable lengths for the weights
        """
        n_active_f, n_active_c = getattr(
            self.regularization_mesh, "aveCC2F{}".format(self.orientation)
        ).shape
        return [(n_active_f,), (n_active_c,)]

    @property
    def cell_gradient(self):
        """Cell gradient operator"""
        if getattr(self, "_cell_gradient", None) is None:
            self._cell_gradient = getattr(
                self.regularization_mesh, "cell_gradient_{}".format(self.orientation)
            )
        return self._cell_gradient

    @property
    def reference_model_in_smooth(self) -> bool:
        """
        Use the reference model in the model gradient penalties.
        """
        return self._reference_model_in_smooth

    @reference_model_in_smooth.setter
    def reference_model_in_smooth(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(
                "'reference_model_in_smooth must be of type 'bool'. "
                f"Value of type {type(value)} provided."
            )
        self._reference_model_in_smooth = value

    def _delta_m(self, m):
        if self.reference_model is None or not self.reference_model_in_smooth:
            return m
        return m - self.reference_model

    @property
    def _multiplier_pair(self):
        return f"alpha_{self.orientation}"

    def f_m(self, m):
        """
        Model gradient
        """
        dfm_dl = self.mapping * self._delta_m(m)

        if self.units is not None and self.units.lower() == "radian":
            return (
                utils.mat_utils.coterminal(self.cell_gradient.sign() @ dfm_dl)
                / self._cell_distances
            )
        return self.cell_gradient @ dfm_dl

    def f_m_deriv(self, m) -> sp.csr_matrix:
        """
        Derivative of the model gradient
        """
        return self.cell_gradient @ self.mapping.deriv(self._delta_m(m))

    @property
    def W(self):
        """
        Weighting matrix that takes the volumes, free weights, fixed weights and
        length scales of the difference operator (normalized optional).
        """
        if getattr(self, "_W", None) is None:
            average_cell_2_face = getattr(
                self.regularization_mesh, "aveCC2F{}".format(self.orientation)
            )
            weights = 1.0
            for values in self._weights.values():
                if values.shape[0] == self.regularization_mesh.nC:
                    values = average_cell_2_face * values
                weights *= values
            self._W = utils.sdiag(weights**0.5)
        return self._W

    @property
    def _cell_distances(self):
        """
        Distances between cell centers for the cell center difference.
        """
        return getattr(self.regularization_mesh, f"cell_distances_{self.orientation}")

    @property
    def orientation(self):
        return self._orientation


class SmoothnessSecondOrder(SmoothnessFirstOrder):
    """
    This base class regularizes on the second
    spatial derivative, optionally normalized by the base cell size.

    **Optional Inputs**

    :param discretize.base.BaseMesh mesh: SimPEG mesh
    :param int nP: number of parameters
    :param IdentityMap mapping: regularization mapping, takes the model from
        model space to the space you want to regularize in
    :param numpy.ndarray reference_model: reference model
    :param numpy.ndarray active_cells: active cell indices for reducing the
        size of differential operators in the definition of a regularization mesh
    :param numpy.ndarray weights: cell weights
    :param bool reference_model_in_smooth: include the reference model in the
        smoothness computation? (eg. look at Deriv of m (False) or Deriv of
        (m-reference_model) (True))
    :param numpy.ndarray weights: vector of cell weights (applied in all terms)
    """

    def f_m(self, m):
        """
        Second model derivative
        """
        dfm_dl = self.mapping * self._delta_m(m)

        if self.units is not None and self.units.lower() == "radian":
            return self.cell_gradient.T @ (
                utils.mat_utils.coterminal(self.cell_gradient.sign() @ dfm_dl)
                / self.length_scales
            )

        dfm_dl2 = self.cell_gradient @ dfm_dl

        return self.cell_gradient.T @ dfm_dl2

    def f_m_deriv(self, m) -> sp.csr_matrix:
        """
        Derivative of the second model residual
        """
        return (
            self.cell_gradient.T
            @ self.cell_gradient
            @ self.mapping.deriv(self._delta_m(m))
        )

    @property
    def W(self):
        """
        Weighting matrix
        """
        if getattr(self, "_W", None) is None:
            weights = np.prod(list(self._weights.values()), axis=0)
            self._W = utils.sdiag(weights**0.5)

        return self._W

    @property
    def _multiplier_pair(self):
        return f"alpha_{self.orientation}{self.orientation}"


###############################################################################
#                                                                             #
#                        Base Combo Regularization                            #
#                                                                             #
###############################################################################


class WeightedLeastSquares(ComboObjectiveFunction):
    r"""Weighted least squares measure on model smallness and smoothness.

    L2 regularization with both smallness and smoothness (first order
    derivative) contributions.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
        The mesh on which the model parameters are defined. This is used
        for constructing difference operators for the smoothness terms.
    active_cells : array_like of bool or int, optional
        List of active cell indices, or a `mesh.n_cells` boolean array
        describing active cells.
    alpha_s : float, optional
        Smallness weight
    alpha_x, alpha_y, alpha_z : float or None, optional
        First order smoothness weights for the respective dimensions.
        `None` implies setting these weights using the `length_scale`
        parameters.
    alpha_xx, alpha_yy, alpha_zz : float, optional
        Second order smoothness weights for the respective dimensions.
    length_scale_x, length_scale_y, length_scale_z : float, optional
        First order smoothness length scales for the respective dimensions.
    mapping : SimPEG.maps.IdentityMap, optional
        A mapping to apply to the model before regularization.
    reference_model : array_like, optional
    reference_model_in_smooth : bool, optional
        Whether to include the reference model in the smoothness terms.
    weights : None, array_like, or dict or array_like, optional
        User defined weights. It is recommended to interact with weights using
        the `get_weights`, `set_weights` functionality.

    Notes
    -----
    The function defined here approximates:

    .. math::
        \phi_m(\mathbf{m}) = \alpha_s \| W_s (\mathbf{m} - \mathbf{m_{ref}} ) \|^2
        + \alpha_x \| W_x \frac{\partial}{\partial x} (\mathbf{m} - \mathbf{m_{ref}} ) \|^2
        + \alpha_y \| W_y \frac{\partial}{\partial y} (\mathbf{m} - \mathbf{m_{ref}} ) \|^2
        + \alpha_z \| W_z \frac{\partial}{\partial z} (\mathbf{m} - \mathbf{m_{ref}} ) \|^2

    Note if the key word argument `reference_model_in_smooth` is False, then mref is not
    included in the smoothness contribution.

    If length scales are used to set the smoothness weights, alphas are respectively set internally using:
    >>> alpha_x = (length_scale_x * min(mesh.edge_lengths)) ** 2
    """

    _model = None

    def __init__(
        self,
        mesh,
        active_cells=None,
        alpha_s=1.0,
        alpha_x=None,
        alpha_y=None,
        alpha_z=None,
        alpha_xx=0.0,
        alpha_yy=0.0,
        alpha_zz=0.0,
        length_scale_x=None,
        length_scale_y=None,
        length_scale_z=None,
        mapping=None,
        reference_model=None,
        reference_model_in_smooth=False,
        weights=None,
        **kwargs,
    ):
        if isinstance(mesh, BaseMesh):
            mesh = RegularizationMesh(mesh)

        if not isinstance(mesh, RegularizationMesh):
            TypeError(
                f"'regularization_mesh' must be of type {RegularizationMesh} or {BaseMesh}. "
                f"Value of type {type(mesh)} provided."
            )
        self._regularization_mesh = mesh
        if active_cells is not None:
            self._regularization_mesh.active_cells = active_cells

        self.alpha_s = alpha_s
        if alpha_x is not None:
            if length_scale_x is not None:
                raise ValueError(
                    "Attempted to set both alpha_x and length_scale_x at the same time. Please "
                    "use only one of them"
                )
            self.alpha_x = alpha_x
        else:
            self.length_scale_x = length_scale_x

        if alpha_y is not None:
            if length_scale_y is not None:
                raise ValueError(
                    "Attempted to set both alpha_y and length_scale_y at the same time. Please "
                    "use only one of them"
                )
            self.alpha_y = alpha_y
        else:
            self.length_scale_y = length_scale_y

        if alpha_z is not None:
            if length_scale_z is not None:
                raise ValueError(
                    "Attempted to set both alpha_z and length_scale_z at the same time. Please "
                    "use only one of them"
                )
            self.alpha_z = alpha_z
        else:
            self.length_scale_z = length_scale_z

        # do this to allow child classes to also pass a list of objfcts to this constructor
        if "objfcts" not in kwargs:
            objfcts = [
                Smallness(mesh=self.regularization_mesh),
                SmoothnessFirstOrder(mesh=self.regularization_mesh, orientation="x"),
                SmoothnessSecondOrder(mesh=self.regularization_mesh, orientation="x"),
            ]

            if mesh.dim > 1:
                objfcts.extend(
                    [
                        SmoothnessFirstOrder(
                            mesh=self.regularization_mesh, orientation="y"
                        ),
                        SmoothnessSecondOrder(
                            mesh=self.regularization_mesh, orientation="y"
                        ),
                    ]
                )

            if mesh.dim > 2:
                objfcts.extend(
                    [
                        SmoothnessFirstOrder(
                            mesh=self.regularization_mesh, orientation="z"
                        ),
                        SmoothnessSecondOrder(
                            mesh=self.regularization_mesh, orientation="z"
                        ),
                    ]
                )
        else:
            objfcts = kwargs.pop("objfcts")
        super().__init__(objfcts=objfcts, **kwargs)

        for fun in objfcts:
            fun.parent = self

        self.mapping = mapping
        self.reference_model = reference_model
        self.reference_model_in_smooth = reference_model_in_smooth
        self.alpha_xx = alpha_xx
        self.alpha_yy = alpha_yy
        self.alpha_zz = alpha_zz
        if weights is not None:
            if not isinstance(weights, dict):
                weights = {"user_weights": weights}
            self.set_weights(**weights)

    def set_weights(self, **weights):
        """Update weights in children objective functions"""
        for fct in self.objfcts:
            fct.set_weights(**weights)

    def remove_weights(self, key):
        """removes weights in children objective functions"""
        for fct in self.objfcts:
            fct.remove_weights(key)

    @property
    def cell_weights(self):
        # All of the objective functions should have the same weights,
        # so just grab the one from smallness here, which should also
        # trigger the deprecation warning
        return self.objfcts[0].cell_weights

    @cell_weights.setter
    def cell_weights(self, value):
        warnings.warn(
            "cell_weights are deprecated please access weights using the `set_weights`,"
            " `get_weights`, and `remove_weights` functionality. This will be removed in 0.19.0",
            FutureWarning,
        )
        self.set_weights(cell_weights=value)

    @property
    def alpha_s(self):
        """smallness weight"""
        return self._alpha_s

    @alpha_s.setter
    def alpha_s(self, value):
        if value is None:
            value = 1.0
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_s must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_s must be non-negative, not {value}")
        self._alpha_s = value

    @property
    def alpha_x(self):
        """weight for the first x-derivative"""
        return self._alpha_x

    @alpha_x.setter
    def alpha_x(self, value):
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_x must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_x must be non-negative, not {value}")
        self._alpha_x = value

    @property
    def alpha_y(self):
        """weight for the first y-derivative"""
        return self._alpha_y

    @alpha_y.setter
    def alpha_y(self, value):
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_y must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_y must be non-negative, not {value}")
        self._alpha_y = value

    @property
    def alpha_z(self):
        """weight for the first z-derivative"""
        return self._alpha_z

    @alpha_z.setter
    def alpha_z(self, value):
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_z must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_z must be non-negative, not {value}")
        self._alpha_z = value

    @property
    def alpha_xx(self):
        """weight for the second x-derivative"""
        return self._alpha_xx

    @alpha_xx.setter
    def alpha_xx(self, value):
        if value is None:
            value = (self.length_scale_x * self.regularization_mesh.base_length) ** 4.0
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_xx must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_xx must be non-negative, not {value}")
        self._alpha_xx = value

    @property
    def alpha_yy(self):
        """weight for the second y-derivative"""
        return self._alpha_yy

    @alpha_yy.setter
    def alpha_yy(self, value):
        if value is None:
            value = (self.length_scale_y * self.regularization_mesh.base_length) ** 4.0
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_yy must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_yy must be non-negative, not {value}")
        self._alpha_yy = value

    @property
    def alpha_zz(self):
        """weight for the second z-derivative"""
        return self._alpha_zz

    @alpha_zz.setter
    def alpha_zz(self, value):
        if value is None:
            value = (self.length_scale_z * self.regularization_mesh.base_length) ** 4.0
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError(f"alpha_zz must be a real number, saw type{type(value)}")
        if value < 0:
            raise ValueError(f"alpha_zz must be non-negative, not {value}")
        self._alpha_zz = value

    @property
    def length_scale_x(self):
        """Constant multiplier of the base length scale on model gradients along x."""
        return np.sqrt(self.alpha_x) / self.regularization_mesh.base_length

    @length_scale_x.setter
    def length_scale_x(self, value: float):
        if value is None:
            value = 1.0
        try:
            value = float(value)
        except (TypeError, ValueError):
            raise TypeError(
                f"length_scale_x must be a real number, saw type{type(value)}"
            )
        self.alpha_x = (value * self.regularization_mesh.base_length) ** 2

    @property
    def length_scale_y(self):
        """Constant multiplier of the base length scale on model gradients along y."""
        return np.sqrt(self.alpha_y) / self.regularization_mesh.base_length

    @length_scale_y.setter
    def length_scale_y(self, value: float):
        if value is None:
            value = 1.0
        try:
            value = float(value)
        except (TypeError, ValueError):
            raise TypeError(
                f"length_scale_y must be a real number, saw type{type(value)}"
            )
        self.alpha_y = (value * self.regularization_mesh.base_length) ** 2

    @property
    def length_scale_z(self):
        """Constant multiplier of the base length scale on model gradients along z."""
        return np.sqrt(self.alpha_z) / self.regularization_mesh.base_length

    @length_scale_z.setter
    def length_scale_z(self, value: float):
        if value is None:
            value = 1.0
        try:
            value = float(value)
        except (TypeError, ValueError):
            raise TypeError(
                f"length_scale_z must be a real number, saw type{type(value)}"
            )
        self.alpha_z = (value * self.regularization_mesh.base_length) ** 2

    @property
    def reference_model_in_smooth(self) -> bool:
        """
        Use the reference model in the model gradient penalties.
        """
        return self._reference_model_in_smooth

    @reference_model_in_smooth.setter
    def reference_model_in_smooth(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(
                "'reference_model_in_smooth must be of type 'bool'. "
                f"Value of type {type(value)} provided."
            )
        self._reference_model_in_smooth = value
        for fct in self.objfcts:
            if getattr(fct, "reference_model_in_smooth", None) is not None:
                fct.reference_model_in_smooth = value

    # Other properties and methods
    @property
    def nP(self):
        """
        number of model parameters
        """
        if getattr(self, "mapping", None) is not None and self.mapping.nP != "*":
            return self.mapping.nP
        elif (
            getattr(self, "_regularization_mesh", None) is not None
            and self.regularization_mesh.nC != "*"
        ):
            return self.regularization_mesh.nC
        else:
            return "*"

    @property
    def _nC_residual(self):
        """
        Shape of the residual
        """

        nC = getattr(self.regularization_mesh, "nC", None)
        mapping = getattr(self, "_mapping", None)

        if mapping is not None and mapping.shape[1] != "*":
            return self.mapping.shape[1]
        elif nC != "*" and nC is not None:
            return self.regularization_mesh.nC
        else:
            return self.nP

    def _delta_m(self, m):
        if self.reference_model is None:
            return m
        return m - self.reference_model

    @property
    def multipliers(self):
        """
        Factors that multiply the objective functions that are summed together
        to build to composite regularization
        """
        return [getattr(self, objfct._multiplier_pair) for objfct in self.objfcts]

    @property
    def active_cells(self) -> np.ndarray:
        """Indices of active cells in the mesh"""
        return self.regularization_mesh.active_cells

    @active_cells.setter
    def active_cells(self, values: np.ndarray):
        self.regularization_mesh.active_cells = values
        active_cells = self.regularization_mesh.active_cells
        # notify the objtecive functions that the active_cells changed
        for objfct in self.objfcts:
            objfct.active_cells = active_cells

    indActive = deprecate_property(
        active_cells,
        "indActive",
        "active_cells",
        "0.19.0",
        error=False,
        future_warn=True,
    )

    @property
    def reference_model(self) -> np.ndarray:
        """Reference physical property model"""
        return self._reference_model

    @reference_model.setter
    def reference_model(self, values: np.ndarray | float):
        if isinstance(values, float):
            values = np.ones(self._nC_residual) * values

        for fct in self.objfcts:
            fct.reference_model = values

        self._reference_model = values

    mref = deprecate_property(
        reference_model,
        "mref",
        "reference_model",
        "0.19.0",
        future_warn=True,
        error=False,
    )

    @property
    def model(self) -> np.ndarray:
        """Physical property model"""
        return self._model

    @model.setter
    def model(self, values: np.ndarray | float):
        if isinstance(values, float):
            values = np.ones(self._nC_residual) * values

        for fct in self.objfcts:
            fct.model = values

        self._model = values

    @property
    def units(self) -> str:
        """Specify the model units. Special care given to 'radian' values"""
        return self._units

    @units.setter
    def units(self, units: str | None):
        if units is not None and not isinstance(units, str):
            raise TypeError(
                f"'units' must be None or type str. "
                f"Value of type {type(units)} provided."
            )
        for fct in self.objfcts:
            fct.units = units
        self._units = units

    @property
    def regularization_mesh(self) -> RegularizationMesh:
        """Regularization mesh"""
        return self._regularization_mesh

    @property
    def mapping(self) -> maps.IdentityMap:
        """Mapping applied to the model values"""
        return self._mapping

    @mapping.setter
    def mapping(self, mapping: maps.IdentityMap):
        if mapping is None:
            mapping = maps.IdentityMap(nP=self._nC_residual)

        if not isinstance(mapping, maps.IdentityMap):
            raise TypeError(
                f"'mapping' must be of type {maps.IdentityMap}. "
                f"Value of type {type(mapping)} provided."
            )
        self._mapping = mapping

        for fct in self.objfcts:
            fct.mapping = mapping


###############################################################################
#                                                                             #
#                        Base Coupling Regularization                         #
#                                                                             #
###############################################################################
class BaseSimilarityMeasure(BaseRegularization):

    """
    Base class for the similarity term in joint inversions. Inherit this for building
    your own similarity term.  The BaseSimilarityMeasure assumes two different
    geophysical models through one similarity term. However, if you wish
    to combine more than two models, e.g., 3 models,
    you may want to add a total of three coupling terms:

    e.g., lambda1*(m1, m2) + lambda2*(m1, m3) + lambda3*(m2, m3)

    where, lambdas are weights for coupling terms. m1, m2 and m3 indicate
    three different models.
    """

    def __init__(self, mesh, wire_map, **kwargs):
        super().__init__(mesh, **kwargs)
        self.wire_map = wire_map

    @property
    def wire_map(self):
        return self._wire_map

    @wire_map.setter
    def wire_map(self, wires: list[maps.IdentityMap]):
        try:
            m1, m2 = wires  # Assume a map has been passed for each model.
        except ValueError:
            ValueError("Wire map must have two model mappings")

        if m1.shape[0] != m2.shape[0]:
            raise ValueError(
                f"All models must be the same size! Got {m1.shape[0]} and {m2.shape[0]}"
            )
        self._wire_map = wires

    @property
    def wire_map_deriv(self):
        if getattr(self, "_wire_map_deriv", None) is None:
            self._wire_map_deriv = sp.vstack([wire.P for wire in self.wire_map])
        return self._wire_map_deriv

    @property
    def nP(self):
        """
        number of model parameters
        """
        return self.wire_map[0].nP

    def deriv(self, model):
        """
        First derivative of the coupling term with respect to individual models.
        Returns an array of dimensions [k*M,1],
        k: number of models we are inverting for.
        M: number of cells in each model.

        """
        raise NotImplementedError(
            "The method deriv has not been implemented for {}".format(
                self.__class__.__name__
            )
        )

    def deriv2(self, model, v=None):
        """
        Second derivative of the coupling term with respect to individual models.
        Returns either an array of dimensions [k*M,1] (v is not None), or
        sparse matrix of dimensions [k*M, k*M] (v is None).
        k: number of models we are inverting for.
        M: number of cells in each model.

        """
        raise NotImplementedError(
            "The method _deriv2 has not been implemented for {}".format(
                self.__class__.__name__
            )
        )

    @property
    def _nC_residual(self):
        """
        Shape of the residual
        """
        return self.wire_map[0].nP

    def __call__(self, model):
        """Returns the computed value of the coupling term."""
        raise NotImplementedError(
            "The method __call__ has not been implemented for {}".format(
                self.__class__.__name__
            )
        )



import numpy as np
from .... import survey
from ....utils import Zero, validate_ndarray_with_shape


class BaseSrc(survey.BaseSrc):
    """Base DC/IP source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.resistivity.receivers.BaseRx
        A list of DC/IP receivers
    location : (n_source, dim) numpy.ndarray
        Source locations
    current : float or numpy.ndarray, default: 1.0
        Current amplitude [A]
    """

    _q = None

    def __init__(self, receiver_list, location, current=1.0, **kwargs):
        super().__init__(receiver_list=receiver_list, **kwargs)
        self.location = location
        self.current = current

    @property
    def location(self):
        """locations of the source electrodes

        Returns
        -------
        (n, dim) numpy.ndarray
            Locations of the source electrodes
        """
        return self._location

    @location.setter
    def location(self, other):
        self._location = validate_ndarray_with_shape(
            "location", other, shape=("*", "*"), dtype=float
        )

    @property
    def current(self):
        """Amplitudes of the source currents

        Returns
        -------
        (n_source) numpy.ndarray
            Amplitudes of the source currents
        """
        return self._current

    @current.setter
    def current(self, other):
        other = validate_ndarray_with_shape("current", other, shape=("*",), dtype=float)
        if len(other) > 1 and len(other) != self.location.shape[0]:
            raise ValueError(
                "Current must be constant or equal to the number of specified source locations."
                f" saw {len(other)} current sources and {self.location.shape[0]} locations."
            )
        self._current = other

    def eval(self, sim):  # noqa: A003
        """Discretize sources to mesh

        Parameters
        ----------
        sim : SimPEG.base.BaseElectricalPDESimulation
            The static electromagnetic simulation

        Returns
        -------
        numpy.ndarray
            The right-hand sides corresponding to the sources
        """
        if sim._formulation == "HJ":
            inds = sim.mesh.closest_points_index(self.location, grid_loc="CC")
            q = np.zeros(sim.mesh.nC)
            q[inds] = self.current
        elif sim._formulation == "EB":
            loc = self.location
            cur = self.current
            interpolation_matrix = sim.mesh.get_interpolation_matrix(
                loc, location_type="N"
            ).toarray()
            q = np.sum(cur[:, np.newaxis] * interpolation_matrix, axis=0)
        return q

    def evalDeriv(self, sim):
        """Returns the derivative of the source term with respect to the model.

        This is zero.

        Parameters
        ----------
        sim : SimPEG.base.BaseElectricalPDESimulation
            The static electromagnetic simulation

        Returns
        -------
        discretize.utils.Zero
        """
        return Zero()


class Multipole(BaseSrc):
    """
    Generic Multipole Source
    """

    @property
    def location_a(self):
        """Locations of the A electrodes

        Returns
        -------
        (n, dim) numpy.ndarray
            Locations of the A electrodes
        """
        return self.location

    @property
    def location_b(self):
        """Location of the B electrodes

        Returns
        -------
        (n, dim) numpy.ndarray
            Locations of the B electrodes
        """
        return np.full_like(self.location, np.nan)


class Dipole(BaseSrc):
    """Dipole source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.resistivity.receivers.BaseRx
        A list of DC/IP receivers
    location_a : (n_source, dim) numpy.array_like
        A electrode locations; remember to set 'location_b' keyword argument to define N electrode locations.
    location_b : (n_source, dim) numpy.array_like
        B electrode locations; remember to set 'location_a' keyword argument to define M electrode locations.
    location : list or tuple of length 2 of numpy.array_like
        A and B electrode locations. In this case, do not set the 'location_a' and 'location_b'
        keyword arguments. And we supply a list or tuple of the form [location_a, location_b].
    """

    def __init__(
        self,
        receiver_list,
        location_a=None,
        location_b=None,
        location=None,
        **kwargs,
    ):
        if "current" in kwargs.keys():
            value = kwargs.pop("current")
            current = [value, -value]
        else:
            current = [1.0, -1.0]

        # if location_a set, then use location_a, location_b
        if location_a is not None:
            if location_b is None:
                raise ValueError(
                    "For a dipole source both location_a and location_b " "must be set"
                )

            if location is not None:
                raise ValueError(
                    "Cannot set both location and location_a, location_b. "
                    "Please provide either location=(location_a, location_b) "
                    "or both location_a=location_a, location_b=location_b"
                )

            location = [location_a, location_b]

        elif location is not None:
            if len(location) != 2:
                raise ValueError(
                    "location must be a list or tuple of length 2: "
                    "[location_a, location_b]. The input location has "
                    f"length {len(location)}"
                )

        # instantiate
        super().__init__(
            receiver_list=receiver_list, location=location, current=current, **kwargs
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(" f"a: {self.location_a}; b: {self.location_b})"
        )

    @property
    def location_a(self):
        """Locations of the A-electrodes

        Returns
        -------
        (n_source, dim) numpy.ndarray
            Locations of the A-electrodes
        """
        return self.location[0]

    @property
    def location_b(self):
        """Locations of the B-electrodes

        Returns
        -------
        (n_source, dim) numpy.ndarray
            Locations of the B-electrodes
        """
        return self.location[1]


class Pole(BaseSrc):
    """Pole source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.resistivity.receivers.BaseRx
        A list of DC/IP receivers
    location : (n_source, dim) numpy.ndarray
        Electrode locations
    """

    def __init__(self, receiver_list, location=None, **kwargs):
        super().__init__(receiver_list=receiver_list, location=location, **kwargs)
        if len(self.location) != 1:
            raise ValueError(
                f"Pole sources only have a single location, not {len(self.location)}"
            )

    @property
    def location_a(self):
        """Locations of the A-electrodes

        Returns
        -------
        (n_source, dim) numpy.ndarray
            Locations of the A-electrodes
        """
        return self.location[0]

    @property
    def location_b(self):
        """Locations of the B-electrodes

        Returns
        -------
        (n_source, dim) numpy.ndarray of ``numpy.nan``
            Locations of the B-electrodes
        """
        return np.full_like(self.location[0], np.nan)



import numpy as np
from discretize.utils import mkvc
import warnings
from ..code_utils import validate_string, validate_type


########################################################################################
#                  DIRECT CURRENT RESISTIVITY AND INDUCED POLARIZATION
########################################################################################


def read_dcip_xyz(
    file_name,
    data_type,
    a_headers=("XA", "YA", "ZA"),
    b_headers=("XB", "YB", "ZB"),
    m_headers=("XM", "YM", "ZM"),
    n_headers=("XN", "YN", "ZN"),
    data_header=None,
    uncertainties_header=None,
    dict_headers=None,
    is_surface_data=False,
):
    """Read 2D or 3D DC/IP data from XYZ-formatted file.

    This function loads DC and/or IP data from XYZ-formatted data files into the
    SimPEG framework; i.e. each row defines the data for the unique electrode
    locations provided. This function is versatile enough to load 2D or 3D data.
    The data file may include elevations for the electrodes or be surface formatted.
    Columns containing data which are not defined as part of a
    :class:`SimPEG.data.Data` object may be loaded and output to a dictionary.

    Parameters
    ----------
    file_name : str
        Path to the data file
    data_type : {"volt", "apparent_resistivity", "apparent_chargeability"}
        Type of data being loaded.
    a_headers : list or tuple of str
        A list or tuple of strings providing the headers of the A-electrode location columns;
        i.e. the X (, Y and Z) columns.
    b_headers : list or tuple of str
        A list or tuple of strings providing the headers of the B-electrode location columns;
        i.e. the X (, Y and Z) columns.
    m_headers : list or tuple of str
        A list or tuple of strings providing the headers of the M-electrode location columns;
        i.e. the X (, Y and Z) columns.
    n_headers : list or tuple of str
        A list or tuple of strings providing the headers of the N-electrode location columns;
        i.e. the X (, Y and Z) columns.
    data_header : str
        The header for the data column
    uncertainties_header : str
        The header for the column containing the data uncertainties
    dict_headers : list or tuple of str
        Provide the hearders for additional data columns that you would like to load and output
        to a dictionary.
    is_surface_data : bool
        If ``True``, we assume electrode elevations are not supplied. That is, the header lists
        for ``a``, ``b``, ``m`` and ``n`` electrode locations do not have headers for
        elevation columns.

    Returns
    -------
    SimPEG.data.Data
        DC or IP data. The survey attribute associated with the data object will be an
        instance of :class:`SimPEG.electromagnetics.static.resistivity.survey.Survey`
        or :class:`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
    dict
        If additional columns are loaded and output to a dictionary using the keyward argument
        `dict_headers`, the output of this function has the form `(out_data, out_dict)`.
    """
    data_type = validate_string(
        "data_type",
        data_type,
        [
            "volt",
            "apparent_resistivity",
            "apparent_chargeability",
        ],
    )

    # Prevent circular import
    from ...electromagnetics.static.utils import generate_survey_from_abmn_locations
    from ...data import Data

    # Load file headers
    FID = open(file_name, "r")
    file_headers = FID.readline()
    FID.close()
    file_headers = file_headers.split()

    # Find indices of columns being loaded
    out_headers = list(a_headers) + list(b_headers) + list(m_headers) + list(n_headers)

    has_data = False
    has_uncert = False
    has_dict = False
    non_location_columns = 0
    if data_header is not None:
        out_headers.append(data_header)
        has_data = True
        non_location_columns += 1
    if uncertainties_header is not None:
        out_headers.append(uncertainties_header)
        has_uncert = True
        non_location_columns += 1
    if dict_headers is not None:
        out_headers += dict_headers
        has_dict = True
        non_location_columns += len(dict_headers)

    col_indices = []
    for h in out_headers:
        col_indices.append(file_headers.index(h))

    # Load specified columns of data
    data_array = np.loadtxt(file_name, comments="!", skiprows=1, usecols=col_indices)
    n_rows = np.shape(data_array)[0]
    num_location_columns = np.shape(data_array)[1] - non_location_columns

    # Extract columns for electrode locations
    if num_location_columns == 12:
        a_cols = [0, 1, 2]
        b_cols = [3, 4, 5]
        m_cols = [6, 7, 8]
        n_cols = [9, 10, 11]
    # 2D survey data or surface 3D data
    elif num_location_columns == 8:
        a_cols = [0, 1]
        b_cols = [2, 3]
        m_cols = [4, 5]
        n_cols = [6, 7]
    # 2D surface data
    else:
        a_cols = [1]
        b_cols = [2]
        m_cols = [3]
        n_cols = [4]

    # Extract electrode locations
    if is_surface_data:
        dummy_elevation = 9999  # Taller than mount Everest
        locations_a = np.c_[data_array[:, a_cols], dummy_elevation * np.ones(n_rows)]
        locations_b = np.c_[data_array[:, b_cols], dummy_elevation * np.ones(n_rows)]
        locations_m = np.c_[data_array[:, m_cols], dummy_elevation * np.ones(n_rows)]
        locations_n = np.c_[data_array[:, n_cols], dummy_elevation * np.ones(n_rows)]
        warnings.warn(
            "Loaded data are in surface format. Elevations automatically set to 9999 m. "
            "Use the project_to_discretized_topography method of the survey to project "
            "electrode locations to the discretized surface."
        )
    else:
        locations_a = data_array[:, a_cols]
        locations_b = data_array[:, b_cols]
        locations_m = data_array[:, m_cols]
        locations_n = data_array[:, n_cols]

    survey, out_indices = generate_survey_from_abmn_locations(
        locations_a=locations_a,
        locations_b=locations_b,
        locations_m=locations_m,
        locations_n=locations_n,
        data_type=data_type,
        output_sorting=True,
    )

    data_object = Data(survey)

    # Sort and organize all data columns
    if has_data:
        data_object.dobs = data_array[out_indices, file_headers.index(data_header)]

    # Sort and organize all data columns
    if has_uncert:
        data_object.standard_deviation = data_array[
            out_indices, file_headers.index(uncertainties_header)
        ]

    # Sort and organize all data columns
    if has_dict:
        out_dict = {}
        for h in dict_headers:
            out_dict[h] = data_array[out_indices, file_headers.index(h)]
        return data_object, out_dict

    else:
        return data_object


def read_dcip2d_ubc(file_name, data_type, format_type):
    """Read UBC-GIF DCIP2D formatted survey or data files.

    This method can load survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP2D coding package. For more, see
    the `UBC-GIF DCIP2D online manual <https://dcip2d.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        The file path to the data file
    data_type : {'volt', 'apparent_chargeability', 'secondary_potential'}
        The type of data.
    format_type : {'general', 'surface', 'simple'}
        The format of the formated file.

    Returns
    -------
    SimPEG.data.Data
        A SimPEG data object. The data from the input file is loaded and parsed into
        three attributes of the data object:

        - `survey`: the survey geometry as defined by an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
        - `dobs`: observed/predicted data if present in the data file
        - `standard_deviations`: uncertainties (if observed data file) or apparent resistivities (if predicted data file)

    """

    data_type = validate_string(
        "data_type",
        data_type,
        [
            "volt",
            "apparent_chargeability",
            "secondary_potential",
        ],
    )

    format_type = validate_string(
        "format_type",
        format_type,
        [
            "general",
            "surface",
            "simple",
        ],
    )

    # Prevent circular import
    from ...electromagnetics.static import resistivity as dc
    from ...electromagnetics.static.utils import generate_survey_from_abmn_locations
    from ...data import Data

    # Load file
    obsfile = np.genfromtxt(file_name, delimiter="\n", dtype=str, comments="!")

    # Find starting data
    start_index = 0
    if obsfile[0] == "COMMON_CURRENT":
        start_index = 2
    if data_type != "volt":
        start_index = start_index + 1

    obsfile = obsfile[start_index:]

    # Since SimPEG defines secondary potential from IP as voltage,
    # we must use this type when defining the receivers.
    if data_type == "secondary_potential":
        data_type = "volt"

    # Pre-allocate
    source_list = []
    d = []
    wd = []

    # Flag for z value provided
    is_surface = False
    is_pole_tx = False
    is_pole_rx = False

    if format_type == "simple":
        # Load numeric data into an array
        if data_type == "volt":
            data_array = np.loadtxt(file_name, comments="!", skiprows=start_index)
        else:
            data_array = np.loadtxt(
                file_name, comments={"!", "IPTYPE"}, skiprows=start_index
            )

        n_rows = np.shape(data_array)[0]

        if np.shape(data_array)[1] > 5:
            wd = data_array[:, 5]

        if np.shape(data_array)[1] > 4:
            d = data_array[:, 4]

        # Get ABMN electrode locations
        dummy_elevation = 9999

        locations_a = np.c_[data_array[:, 0], dummy_elevation * np.ones(n_rows)]
        locations_b = np.c_[data_array[:, 1], dummy_elevation * np.ones(n_rows)]
        locations_m = np.c_[data_array[:, 2], dummy_elevation * np.ones(n_rows)]
        locations_n = np.c_[data_array[:, 3], dummy_elevation * np.ones(n_rows)]

        survey, out_indices = generate_survey_from_abmn_locations(
            locations_a=locations_a,
            locations_b=locations_b,
            locations_m=locations_m,
            locations_n=locations_n,
            data_type=data_type,
            output_sorting=True,
        )

        data_out = Data(survey)

        # Sort and organize all data columns
        if len(d) > 0:
            data_out.dobs = d[out_indices]

        # Sort and organize all data columns
        if len(wd) > 0:
            data_out.standard_deviation = wd[out_indices]

        warnings.warn(
            "Loaded data did not have elevations. Elevations automatically set to 9999 m. "
            "Use the project_to_discretized_topography method of the survey to project "
            "electrode locations to the discretized surface."
        )

    else:
        # Countdown for number of obs/tx
        count = 0
        for ii in range(obsfile.shape[0]):
            if not obsfile[ii]:
                continue

            # Extract transmitter ii and the number of receivers
            if count == 0:
                rx = []
                temp = np.fromstring(obsfile[ii], dtype=float, sep=" ").T
                count = int(temp[-1])

                # Check if z value is provided, if False -> 9999
                if len(temp) == 3:
                    # check if pole|dipole
                    if np.allclose(temp[0], temp[1]):
                        tx = np.r_[temp[0], 9999]
                        is_pole_tx = True
                    else:
                        tx = np.r_[temp[0], 9999, temp[1], 9999]
                    is_surface = True

                else:
                    # check if pole|dipole
                    if np.allclose(temp[0:2], temp[2:4]):
                        tx = np.r_[temp[0:2]]
                        is_pole_tx = True
                    else:
                        tx = temp[:-1]

                continue

            # Extract receivers
            temp = np.fromstring(obsfile[ii], dtype=float, sep=" ")

            if is_surface:
                data_column_index = 2

                # Check if Pole Receiver
                if np.allclose(temp[0], temp[1]):
                    is_pole_rx = True
                    rx.append(np.r_[temp[0], 9999])
                else:
                    rx.append(np.r_[temp[0], 9999, temp[1], 9999])

            else:
                data_column_index = 4  # Since dpred for dc has app_res

                # Check if Pole Receiver
                if np.allclose(temp[0:2], temp[2:4]):
                    is_pole_rx = True
                    rx.append(temp[:2])
                else:
                    rx.append(temp[:4])

            # Predicted/observed data
            if len(temp) == data_column_index + 1:
                d.append(temp[data_column_index])

            # Observed data or predicted DC data (since app res column)
            elif len(temp) == data_column_index + 2:
                d.append(temp[data_column_index])
                wd.append(temp[data_column_index + 1])

            count = count - 1

            # Reach the end of transmitter block
            if count == 0:
                rx = np.asarray(rx)
                if is_pole_rx:
                    Rx = dc.receivers.Pole(rx[:, :2], data_type=data_type)
                else:
                    Rx = dc.receivers.Dipole(rx[:, :2], rx[:, 2:], data_type=data_type)
                if is_pole_tx:
                    source_list.append(dc.sources.Pole([Rx], tx[:2]))
                else:
                    source_list.append(dc.sources.Dipole([Rx], tx[:2], tx[2:]))

        survey = dc.survey.Survey(source_list)
        data_out = Data(survey=survey)

        if len(d) > 0:
            data_out.dobs = d

        if len(wd) > 0:
            data_out.standard_deviation = wd

        if is_surface:
            warnings.warn(
                "Loaded data were in surface format. Elevations automatically set to 9999 m. "
                "Use the project_to_discretized_topography method of the survey to project "
                "electrode locations to the discretized surface."
            )

    return data_out


def read_dcip3d_ubc(file_name, data_type):
    """Read UBC-GIF DCIP3D formatted survey or data files.

    This method can load survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP3D coding package. For more, see
    the `UBC-GIF DCIP3D online manual <https://dcip3d.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        The file path to the data file
    data_type : {'volt', 'apparent_chargeability', secondary_potential'}
        Data type.

    Returns
    -------
    SimPEG.data.Data
        A SimPEG data object. The data from the input file is loaded and parsed into
        three attributes of the data object:

        - `survey`: the survey geometry as defined by an instance of :class`SimPEG.electromagnetics.static.resitivity.survey.Survey` or :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
        - `dobs`: observed/predicted data if present in the data file
        - `standard_deviations`: uncertainties (if observed data file) or apparent resistivities (if predicted data file)

    """

    data_type = validate_string(
        "data_type",
        data_type,
        [
            "volt",
            "apparent_chargeability",
            "secondary_potential",
        ],
    )

    # Prevent circular import
    from ...electromagnetics.static import resistivity as dc
    from ...data import Data

    # Load file
    obsfile = np.genfromtxt(file_name, delimiter="\n", dtype=str, comments="!")

    # Pre-allocate
    source_list = []
    d = []
    wd = []

    # Flag for z value provided
    is_surface = False
    is_pole_tx = False
    is_pole_rx = False

    # IP data for dcip3d has a line with a flag we can remove.
    if obsfile[0][0:6] == "IPTYPE":
        obsfile = obsfile[1:]

    # Since SimPEG defines secondary potential from IP as voltage,
    # we must use this type when defining the receivers.
    if data_type == "secondary_potential":
        data_type = "volt"

    # Countdown for number of obs/tx
    count = 0
    for ii in range(obsfile.shape[0]):
        if not obsfile[ii]:
            continue

        # Extract transmitter ii and the number of receivers
        if count == 0:
            rx = []
            temp = np.fromstring(obsfile[ii], dtype=float, sep=" ").T
            count = int(temp[-1])

            # Check if z value is provided, if False -> 9999
            if len(temp) == 5:
                # check if pole|dipole
                if np.allclose(temp[0:2], temp[2:4]):
                    tx = np.r_[temp[0:2], 9999]
                    is_pole_tx = True
                else:
                    tx = np.r_[temp[0:2], 9999, temp[2:4], 9999]
                is_surface = True

            else:
                # check if pole|dipole
                if np.allclose(temp[0:3], temp[3:6]):
                    tx = np.r_[temp[0:3]]
                    is_pole_tx = True
                else:
                    tx = temp[:-1]

            continue

        # Extract receivers
        temp = np.fromstring(obsfile[ii], dtype=float, sep=" ")

        if is_surface:
            data_column_index = 4  # Since dpred for dc has app_res

            # Check if Pole Receiver
            if np.allclose(temp[0:2], temp[2:4]):
                is_pole_rx = True
                rx.append(temp[:2])
            else:
                rx.append(np.r_[temp[0:2], 9999, temp[2:4], 9999])

        else:
            data_column_index = 6  # Since dpred for dc has app_res

            # Check if Pole Receiver
            if np.allclose(temp[0:3], temp[3:6]):
                is_pole_rx = True
                rx.append(temp[:3])
            else:
                rx.append(temp[:6])

        # Predicted/observed data
        if len(temp) == data_column_index + 1:
            d.append(temp[data_column_index])

        # Observed data or predicted DC data (since app res column)
        elif len(temp) == data_column_index + 2:
            d.append(temp[data_column_index])
            wd.append(temp[data_column_index + 1])

        count = count - 1

        # Reach the end of transmitter block
        if count == 0:
            rx = np.asarray(rx)
            if is_pole_rx:
                Rx = dc.receivers.Pole(rx[:, :3], data_type=data_type)
            else:
                Rx = dc.receivers.Dipole(rx[:, :3], rx[:, 3:], data_type=data_type)
            if is_pole_tx:
                source_list.append(dc.sources.Pole([Rx], tx[:3]))
            else:
                source_list.append(dc.sources.Dipole([Rx], tx[:3], tx[3:]))

    survey = dc.survey.Survey(source_list)
    data_out = Data(survey=survey)

    if len(d) > 0:
        data_out.dobs = d

    if len(wd) > 0:
        data_out.standard_deviation = wd

    if is_surface:
        warnings.warn(
            "Loaded data were in surface format. Elevations automatically set to 9999 m. "
            "Use the project_to_discretized_topography method of the survey to project "
            "electrode locations to the discretized surface."
        )

    return data_out


def read_dcipoctree_ubc(file_name, data_type):
    """Read UBC-GIF DCIP OcTree formatted survey or data files.

    This method can load survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP OcTree coding package. For more, see
    the `UBC-GIF DCIP OcTree online manual <https://dcipoctree.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        The file path to the data file
    data_type : {'volt', 'apparent_chargeability', secondary_potential'}
        Data type.

    Returns
    -------
    SimPEG.data.Data
        A SimPEG data object. The data from the input file is loaded and parsed into
        three attributes of the data object:

        - `survey`: the survey geometry as defined by an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
        - `dobs`: observed/predicted data if present in the data file
        - `standard_deviations`: uncertainties (if observed data file) or apparent resistivities (if predicted data file)

    """

    return read_dcip3d_ubc(file_name, data_type)


def write_dcip2d_ubc(
    file_name,
    data_object,
    data_type,
    file_type,
    format_type="general",
    comment_lines=None,
):
    """Write UBC-GIF DCIP2D formatted survey or data files.

    This function can write survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP2D coding package. For more, see
    the `UBC-GIF DCIP2D online manual <https://dcip2d.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        file path for output file
    data_object :
        SimPEG.data.Data object. The `survey` attribute of this data object must be
        an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or
        :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
    data_type : {'volt', 'apparent_chargeability', 'secondary_potential'}
        The type of data.
    file_type : {'survey', 'dpred', 'dobs'}
        Whether to include predicted data/observed data in the file.
    format_type : {'general', 'surface', 'simple'}
        The format of the file to write.
    comment_lines :
        Comment lines printed to beginning of the file
    """

    # Prevent circular import
    from ...electromagnetics.static.resistivity.sources import (
        Pole as PoleSrc,
        Dipole as DipoleSrc,
    )
    from ...electromagnetics.static.resistivity.receivers import (
        Pole as PoleRx,
        Dipole as DipoleRx,
    )

    from ...data import Data

    # Validate inputs
    data_object = validate_type("data_object", data_object, Data, cast=False)

    data_type = validate_string(
        "data_type",
        data_type,
        [
            "volt",
            "apparent_chargeability",
            "secondary_potential",
        ],
    )

    file_type = validate_string(
        "file_type",
        file_type,
        [
            "survey",
            "dpred",
            "dobs",
        ],
    )

    format_type = validate_string(
        "format_type",
        format_type,
        [
            "general",
            "surface",
            "simple",
        ],
    )

    # Write comments and IP type (if applicable)
    with open(file_name, "w") as fid:
        if format_type != "simple":
            fid.write("COMMON_CURRENT\n")

        fid.write(f"! {format_type} FORMAT\n")

        if comment_lines is not None and len(comment_lines) > 0:
            # ensure comment_lines ends with a new line character
            if comment_lines[0] != "!":
                comment_lines = "! " + comment_lines
            if comment_lines[-1] != "\n":
                comment_lines += "\n"
            fid.write(comment_lines)

        if format_type.lower() != "simple":
            fid.write("{:g}\n".format(len(data_object.survey.source_list)))

        # DCIP3D will allow user to choose definition of IP data. DC data has no flag.
        # DCIPoctree IP data is always apparent chargeability.
        if data_type == "apparent_chargeability":
            fid.write("IPTYPE=1\n")
        elif data_type == "secondary_potential":
            fid.write("IPTYPE=2\n")

        if format_type.lower() == "simple":
            out_array = np.c_[
                data_object.survey.locations_a[:, 0],
                data_object.survey.locations_b[:, 0],
                data_object.survey.locations_m[:, 0],
                data_object.survey.locations_n[:, 0],
            ]

            if file_type != "survey":
                out_array = np.c_[out_array, data_object.dobs]

            if file_type == "dobs":
                out_array = np.c_[out_array, data_object.standard_deviation]

            np.savetxt(fid, out_array, fmt="%e", delimiter="    ")

        else:
            # Index deciding if z locations are written
            if format_type == "surface":
                end_index = 1
            elif format_type == "general":
                end_index = 2

            # Loop over all sources
            count = 0
            for src in data_object.survey.source_list:
                # Write Source
                nD = src.nD

                if isinstance(src, PoleSrc):
                    tx = src.location[0]
                    tx = np.repeat(np.r_[[tx]], 2, axis=0)
                elif isinstance(src, DipoleSrc):
                    tx = np.c_[src.location]

                fid.writelines("%e " % ii for ii in mkvc(tx[:, 0:end_index].T))
                fid.write(f"{nD}\n")

                # Write receivers
                for rx in src.receiver_list:
                    if isinstance(rx, DipoleRx):
                        M = rx.locations[0][:, 0:end_index]
                        N = rx.locations[1][:, 0:end_index]
                    elif isinstance(rx, PoleRx):
                        M = rx.locations[:, 0:end_index]
                        N = rx.locations[:, 0:end_index]

                    if file_type.lower() != "survey":
                        N = np.c_[N, data_object.dobs[count : count + rx.nD]]

                    if file_type.lower() == "dobs":
                        N = np.c_[
                            N, data_object.standard_deviation[count : count + rx.nD]
                        ]

                    # Write receivers and locations
                    if isinstance(N, np.ndarray):
                        np.savetxt(
                            fid,
                            np.c_[M, N],
                            fmt="%e",
                        )
                    else:
                        raise Exception(
                            """Uncertainities SurveyObject.std should be set.
                            Either float or nunmpy.ndarray is expected, """
                            "not {}".format(type(data_object.relative_error))
                        )

                    fid.write("\n")

                    count += rx.nD


def write_dcip3d_ubc(
    file_name,
    data_object,
    data_type,
    file_type,
    format_type="general",
    comment_lines=None,
):
    """Write UBC-GIF DCIP3D formatted survey or data files.

    This function can write survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP3D coding package. For more, see
    the `UBC-GIF DCIP3D online manual <https://dcip3d.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        file path for output file
    data_object :
        SimPEG.data.Data object. The `survey` attribute of this data object must be
        an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or
        :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
    data_type : {'volt', 'apparent_chargeability', 'secondary_potential'}
    file_type : {'survey', 'dpred', 'dobs'}
    format_type : {'general', 'surface'}
    comment_lines : str
        Comments added to beginning of output file
    """

    # Prevent circular import
    from ...electromagnetics.static.utils.static_utils import (
        apparent_resistivity_from_voltage,
    )
    from ...electromagnetics.static.resistivity.sources import (
        Pole as PoleSrc,
        Dipole as DipoleSrc,
    )
    from ...electromagnetics.static.resistivity.receivers import (
        Pole as PoleRx,
        Dipole as DipoleRx,
    )
    from ...data import Data

    # Validate inputs
    data_object = validate_type("data_object", data_object, Data, cast=False)

    data_type = validate_string(
        "data_type",
        data_type,
        [
            "volt",
            "apparent_chargeability",
            "secondary_potential",
        ],
    )

    file_type = validate_string(
        "file_type",
        file_type,
        [
            "survey",
            "dpred",
            "dobs",
        ],
    )

    format_type = validate_string(
        "format_type",
        format_type,
        [
            "general",
            "surface",
        ],
    )

    # Predicted DC data will automatically contain apparent resistivity column.
    # Here we compute the apparent resistivities and treat it like an uncertainties column.
    if (file_type == "dpred") & (data_type == "volt"):
        data_object.standard_deviation = apparent_resistivity_from_voltage(
            data_object.survey, data_object.dobs
        )
        file_type = "dobs"

    # Write comments and IP type (if applicable)
    with open(file_name, "w") as fid:
        fid.write(f"! {format_type} FORMAT\n")

        if comment_lines is not None and len(comment_lines) > 0:
            if comment_lines[0] != "!":
                comment_lines = "! " + comment_lines
            # ensure comment_lines ends with a new line character
            if comment_lines[-1] != "\n":
                comment_lines += "\n"
            fid.write(comment_lines)

        if data_type == "apparent_chargeability":
            fid.write("IPTYPE=1\n")
        elif data_type == "secondary_potential":
            fid.write("IPTYPE=2\n")

        # Index deciding if z locations are written
        if format_type == "surface":
            end_index = 2
        elif format_type == "general":
            end_index = 3

        # Loop over all sources
        count = 0
        for src in data_object.survey.source_list:
            # Write Source
            nD = src.nD

            if isinstance(src, PoleSrc):
                tx = src.location[0]
                tx = np.repeat(np.r_[[tx]], 2, axis=0)
            elif isinstance(src, DipoleSrc):
                tx = np.c_[src.location]

            fid.writelines("%e " % ii for ii in mkvc(tx[:, 0:end_index].T))
            fid.write(f"{nD}\n")

            # Write receivers
            for rx in src.receiver_list:
                if isinstance(rx, DipoleRx):
                    M = rx.locations[0][:, 0:end_index]
                    N = rx.locations[1][:, 0:end_index]
                elif isinstance(rx, PoleRx):
                    M = rx.locations[:, 0:end_index]
                    N = rx.locations[:, 0:end_index]

                if file_type != "survey":
                    N = np.c_[N, data_object.dobs[count : count + rx.nD]]

                if file_type == "dobs":
                    N = np.c_[N, data_object.standard_deviation[count : count + rx.nD]]

                # Write receivers and locations
                if isinstance(N, np.ndarray):
                    np.savetxt(
                        fid,
                        np.c_[M, N],
                        fmt="%e",
                    )
                else:
                    raise Exception(
                        """Uncertainities SurveyObject.std should be set.
                        Either float or nunmpy.ndarray is expected, """
                        "not {}".format(type(data_object.relative_error))
                    )

                fid.write("\n")

                count += rx.nD


def write_dcipoctree_ubc(
    file_name,
    data_object,
    data_type,
    file_type,
    format_type="general",
    comment_lines="",
):
    """Write UBC-GIF DCIP OcTree formatted survey or data files.

    This function can write survey locations, predicted data or observations
    files formatted for the UBC-GIF DCIP OcTree coding package. For more, see
    the `UBC-GIF DCIP OcTree online manual <https://dcipoctree.readthedocs.io/en/latest/>`__.

    Parameters
    ----------
    file_name : str
        file path for output file
    data_object :
        SimPEG.data.Data object. The `survey` attribute of this data object must be
        an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or
        :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
    data_type : {'volt', 'apparent_chargeability', 'secondary_potential'}
    file_type : {'survey', 'dpred', 'dobs'}
    format_type : {'general', 'surface'}
    comment_lines : str
        Comments added to beginning of output file
    """

    write_dcip3d_ubc(
        file_name,
        data_object,
        data_type,
        file_type,
        format_type=format_type,
        comment_lines=comment_lines,
    )


def write_dcip_xyz(
    file_name, data_object, data_header=None, uncertainties_header=None, out_dict=None
):
    """Write 2D or 3D DC/IP data to XYZ-formatted file.

    This function writes DC and/or IP data from the SimPEG framework to an XYZ-formatted
    data file; i.e. each row in the file defines the data for the unique electrode
    locations provided. This function is versatile enough to write 2D or 3D data.
    The data may include elevations for the electrodes or be surface formatted.

    Parameters
    ----------
    file_name : str
        Path to the file
    data_object : SimPEG.data.Data
        SimPEG.data.Data object. The `survey` attribute of this data object must be
        an instance of :class`SimPEG.electromagnetics.static.resistivity.survey.Survey` or
        :class`SimPEG.electromagnetics.static.induced_polarization.survey.Survey`
    data_header: str
        Header for the data column; i.e. the header for the data defined in the `dobs`
        attibute of the data object. If ``None``, these data are not written to file
    uncertainties_header : str
        Header for the uncertainties column; i.e. the header for the uncertainties defined in
        the `standard_deviations` attibute of the data object. If ``None``, these data are not
        written to file
    out_dict : dict
        A python dictionary containing the names and associated vectors for any additional data
        columns you wish to write to the file; out_dict = {header1: vec1, header2: vec2, ...}.
    """

    out_columns = np.c_[
        data_object.survey.locations_a,
        data_object.survey.locations_b,
        data_object.survey.locations_m,
        data_object.survey.locations_n,
    ]

    # Determine if 2D or 3D survey
    if np.shape(out_columns)[1] == 8:
        out_headers = "XA    ZA    XB    ZB    XM    ZM    XN    ZN"
    else:
        out_headers = (
            "XA    YA    ZA    XB    YB    ZB    XM    YM    ZM    XN    YN    ZN"
        )

    # Append data and uncertainties headers
    if (data_object.dobs is not None) & (data_header is not None):
        out_headers += "    " + data_header
        out_columns = np.c_[out_columns, data_object.dobs]

    if (data_object.standard_deviation is not None) & (
        uncertainties_header is not None
    ):
        out_headers += "    " + uncertainties_header
        out_columns = np.c_[out_columns, data_object.standard_deviation]

    # Append additional columns from dictionary
    if out_dict != None:
        for k in list(out_dict.keys()):
            out_headers += "    " + k
            out_columns = np.c_[out_columns, out_dict[k]]

    # Write to file
    np.savetxt(
        file_name,
        out_columns,
        fmt="%.8e",
        delimiter="    ",
        newline="\n",
        header=out_headers,
        comments="",
    )

    print("XYZ file saved to: " + file_name)



from .io_utils_general import read_GOCAD_ts, download
from .io_utils_pf import (
    read_mag3d_ubc,
    write_mag3d_ubc,
    read_grav3d_ubc,
    write_grav3d_ubc,
    read_gg3d_ubc,
    write_gg3d_ubc,
)

from .io_utils_electromagnetics import (
    read_dcip2d_ubc,
    read_dcip3d_ubc,
    read_dcipoctree_ubc,
    read_dcip_xyz,
    write_dcip2d_ubc,
    write_dcip3d_ubc,
    write_dcipoctree_ubc,
    write_dcip_xyz,
)

# Deprecated
from .io_utils_pf import (
    readUBCmagneticsObservations,
    writeUBCmagneticsObservations,
    readUBCgravityObservations,
    writeUBCgravityObservations,
)



import numpy as np

from ... import maps
from ..frequency_domain.sources import BaseFDEMSrc
from ..utils import omega
from .utils.source_utils import homo1DModelSource
import discretize
from discretize.utils import volume_average


#################
#    Sources    #
#################


class Planewave(BaseFDEMSrc):
    """
    Source class for the 1D and pseudo-3D problems.

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.frequency_domain.receivers.BaseRx
        A list of NSEM receivers
    frequency : float
        Source frequency
    """

    # This class is only provided to have a uniquely identifiable "Planewave" class
    pass


# Need to implement such that it works for all dims.
# Rename to be more descriptive
class PlanewaveXYPrimary(Planewave):
    """
    NSEM planewave source for both polarizations (x and y)
    estimated from a single 1D primary models.

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.frequency_domain.receivers.BaseRx
        A list of NSEM receivers
    frequency : float
        Source frequency
    sigma_primary : float, default: ``None``
        Wholespace conductivity for primary field
    """

    _fields_per_source = 2

    def __init__(self, receiver_list, frequency, sigma_primary=None):
        # assert mkvc(self.mesh.h[2].shape,1) == mkvc(sigma1d.shape,1),'The number of values in the 1D background model does not match the number of vertical cells (hz).'
        self.sigma1d = None
        self._sigma_primary = sigma_primary
        super(PlanewaveXYPrimary, self).__init__(receiver_list, frequency)

    def _get_sigmas(self, simulation):
        try:
            return self._sigma1d, self._sigma_p
        except AttributeError:
            # set _sigma 1D
            if self._sigma_primary is None:
                self._sigma_primary = simulation.sigmaPrimary
            # Create 3d_1d mesh like me...
            if simulation.mesh.dim == 3:
                mesh3d = simulation.mesh
                x0 = mesh3d.x0
                hs = [
                    [mesh3d.nodes_x[-1] - x0[0]],
                    [mesh3d.nodes_y[-1] - x0[1]],
                    mesh3d.h[-1],
                ]
                mesh1d = discretize.TensorMesh(hs, x0=x0)
                if len(self._sigma_primary) == mesh3d.nC:
                    # volume average down to 1D mesh
                    self._sigma1d = np.exp(
                        volume_average(mesh3d, mesh1d, np.log(self._sigma_primary))
                    )
                elif len(self._sigma_primary) == mesh1d.nC:
                    self._sigma1d = self._sigma_primary
                else:
                    self._sigma1d = np.ones(mesh1d.nC) * self._sigma_primary
                self._sigma_p = np.exp(
                    volume_average(mesh1d, mesh3d, np.log(self._sigma1d))
                )
            else:
                self._sigma1d = simulation.mesh.reshape(
                    simulation._sigmaPrimary, "CC", "CC", "M"
                )[:]
                self._sigma_p = None
                self.sigma1d = self._sigma1d
            return self._sigma1d, self._sigma_p

    def ePrimary(self, simulation):
        """Primary electric field

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.natural_source_simulation.BaseNSEMSimulation
            A NSEM simulation

        Returns
        -------
        numpy.ndarray
            Primary electric field
        """
        if self._ePrimary is None:
            sigma_1d, _ = self._get_sigmas(simulation)
            self._ePrimary = homo1DModelSource(
                simulation.mesh, self.frequency, sigma_1d
            )
        return self._ePrimary

    def bPrimary(self, simulation):
        """Primary magnetic field

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.frequency_domain.simulation.BaseFDEMSimulation
            A NSEM simulation

        Returns
        -------
        numpy.ndarray
            Primary magnetic field
        """
        # Project ePrimary to bPrimary
        # Satisfies the primary(background) field conditions
        if simulation.mesh.dim == 1:
            C = simulation.mesh.nodal_gradient
        elif simulation.mesh.dim == 3:
            C = simulation.mesh.edge_curl
        bBG_bp = (-C * self.ePrimary(simulation)) * (1 / (1j * omega(self.frequency)))
        return bBG_bp

    def s_e(self, simulation):
        """Electric source term

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.frequency_domain.simulation.BaseFDEMSimulation
            A NSEM simulation

        Returns
        -------
        numpy.ndarray
            Electric source term on mesh.
        """
        e_p = self.ePrimary(simulation)
        # Make mass matrix
        # Note: M(sig) - M(sig_p) = M(sig - sig_p)
        # Need to deal with the edge/face discrepencies between 1d/2d/3d
        if simulation.mesh.dim == 1:
            Map_sigma_p = maps.SurjectVertical1D(simulation.mesh)
            sigma_p = Map_sigma_p._transform(self.sigma1d)
            Mesigma = simulation.mesh.get_face_inner_product(simulation.sigma)
            Mesigma_p = simulation.mesh.get_face_inner_product(sigma_p)
        if simulation.mesh.dim == 2:
            pass
        if simulation.mesh.dim == 3:
            _, sigma_p = self._get_sigmas(simulation)
            Mesigma = simulation.MeSigma
            Mesigma_p = simulation.mesh.get_edge_inner_product(sigma_p)
        return Mesigma * e_p - Mesigma_p * e_p

    def s_eDeriv(self, simulation, v, adjoint=False):
        """Derivative of electric source term with respect to model

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.frequency_domain.simulation.BaseFDEMSimulation
            A NSEM simulation
        v : numpy.ndarray
            A vector
        adjoint : bool, default: ``False``
            If ``True``, perform the adjoint operation

        Returns
        -------
        numpy.ndarray
            Derivative of electric source term on mesh.
        """

        return self.s_eDeriv_m(simulation, v, adjoint)

    def s_eDeriv_m(self, simulation, v, adjoint=False):
        """Derivative of electric source term with respect to model

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.frequency_domain.simulation.BaseFDEMSimulation
            A NSEM simulation
        v : numpy.ndarray
            A vector
        adjoint : bool, default: ``False``
            If ``True``, perform the adjoint operation

        Returns
        -------
        numpy.ndarray
            Derivative of electric source term on mesh.
        """
        # Need to deal with
        if simulation.mesh.dim == 1:
            # Need to use the faceInnerProduct
            ePri = self.ePrimary(simulation)[:, 1]
            return simulation.MfSigmaDeriv(ePri, v, adjoint=adjoint)
        if simulation.mesh.dim == 2:
            raise NotImplementedError("The NSEM 2D simulation is not implemented")
        if simulation.mesh.dim == 3:
            # Need to take the derivative of both u_px and u_py
            # And stack them to be of the correct size
            e_p = self.ePrimary(simulation)
            return simulation.MeSigmaDeriv(e_p, v, adjoint=adjoint)

    S_e = s_e
    S_eDeriv = s_eDeriv



import matplotlib.pyplot as plt
import numpy as np
from scipy.constants import mu_0

from ...utils import omega
from ....utils import validate_type, validate_list_of_types


# Define the default component dictionaries
DEFAULT_COMP_DICT = {
    "xx": {"color": "green", "label": "Imp_xx", "marker": "_", "ls": "None"},
    "xy": {"color": "blue", "label": "Imp_xy", "marker": "_", "ls": "None"},
    "yx": {"color": "red", "label": "Imp_yx", "marker": "_", "ls": "None"},
    "yy": {"color": "yellow", "label": "Imp_yy", "marker": "_", "ls": "None"},
    "zx": {"color": "brown", "label": "Tip_zx", "marker": "_", "ls": "None"},
    "zy": {"color": "purple", "label": "Tip_zy", "marker": "_", "ls": "None"},
}

# Define some hidden attributes
_imp_comps = ["xx", "xy", "yx", "yy"]


def _validate_kwargs(input_dict, compare_dict):
    """
    Function to deal with keyword arguments.

    :param input_dict: matplotlib kwargs dictionary with custom arguments
    :type input_dict: :class: `dict`
    :param compare_dict: matplotlib kwargs of default to use arguments
    :type compare_dict: :class: `dict`
    """
    # Set the default plot kwargs
    for key, val in compare_dict.items():
        # Make sure they aren't already assigned
        if input_dict is None:
            input_dict = compare_dict.copy()
        else:
            if key not in input_dict:
                input_dict[key] = val
    # Return
    return input_dict


class BaseDataNSEMPlots:
    """
    A class container of matplotlib panels for plotting
    NSEM data.

    """

    def __init__(self, fig=None, axes=None, **kwargs):
        super().__init__(**kwargs)
        self.fig = fig
        self.axes = axes

    @property
    def fig(self):
        """Figure for plotting.

        Returns
        -------
        matplotlib.figure.Figure
        """
        return self._fig

    @fig.setter
    def fig(self, value):
        if value is not None:
            value = validate_type("fig", value, plt.Figure, cast=False)
        self._fig = value

    @property
    def axes(self):
        """Figure for plotting.

        Returns
        -------
        list of matplotlib.axes.Axes
        """
        return self._axes

    @axes.setter
    def axes(self, value):
        if value is not None:
            value = validate_list_of_types("axes", value, plt.Axes)
        self._axes = value

    def setup(self):
        """
        Setup up the plot window.

        Should populate the
            self.fig and self.axes properties
        """
        raise NotImplementedError(
            "Is required in subclasses of {}".format(self.__class__)
        )

    def draw(self):
        raise NotImplementedError(
            "Is required in subclasses of {}".format(self.__class__)
        )

    def clear_axes(self):
        """
        Function to clear all of the axes
        """
        for ax in self.axes:
            while len(ax.lines) > 0:
                for line in ax.lines:
                    ax.lines.remove(line)

            while len(ax.collections) > 0:
                for item in ax.collections:
                    item.remove()


class TipperAmplitudeStationPlot(BaseDataNSEMPlots):
    """
    Class for setting up 2 axes figure with:
        tipper amplitudes | tipper phase
        setup.
    """

    def __init__(self):
        super(TipperAmplitudeStationPlot, self).__init__()

    def setup(self):
        """
        Setup a station data plot figure.
        """
        self.fig, axes_temp = plt.subplots(1, 2, sharex=True)
        self.axes = axes_temp.ravel().tolist()
        self.fig.set_size_inches((13.5, 4.0))

        for ax in self.axes:
            ax.set_xscale("log")

        self.axes[0].invert_xaxis()
        self.axes[0].set_yscale("log")
        # Set labels
        self.axes[0].set_xlabel("Frequency [Hz]")
        self.axes[1].set_xlabel("Frequency [Hz]")
        self.axes[0].set_ylabel("Tipper amplitude [V/A]")
        self.axes[1].set_ylim(-180, 180)
        self.axes[1].set_ylabel("Tipper angle [degrees]")

    def draw(self, data_list, location):
        """
        Function to draw on the axes

        :param data_list: List of NSEM data objects to plot.
            Has to be of length >= 1. First item is treat as a
            observed data (Hast to have relative_error and noise_floor)
            assigned) and the others are plotted on top.
        :param location: Location of the station to plot
        """

        axes = self.axes

        # Pop the data from the list
        data = data_list[0]

        # Apparent resistivity
        data.plot_tip_amp(location, ["zx", "zy"], ax=axes[0], errorbars=True)

        # Apparent phase
        data.plot_app_phs(location, ["zx", "zy"], ax=axes[1], errorbars=True)

        # Plot the additional data
        for other_data in data_list[1::]:
            # Need add symbol generation
            dd_kwargs = {
                "zx": {"marker": ".", "ls": "--"},
                "zy": {"marker": ".", "ls": "--"},
            }

            # Apparent resistivity
            other_data.plot_tip_amp(
                location,
                ["zx", "zy"],
                ax=axes[0],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )

            # Apparent phase
            other_data.plot_app_phs(
                location,
                ["zx", "zy"],
                ax=axes[1],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )


class ApparentResPhsStationPlot(BaseDataNSEMPlots):
    """
    Class for setting up 4 axes figure with:

            +---------------------+----------------+
            |apparent resistivity | phase          |
            +---------------------+----------------+
            |impedance amplitudes | impedance phase|
            +---------------------+----------------+

        setup.

    """

    def __init__(self):
        super(ApparentResPhsStationPlot, self).__init__()

    def setup(self):
        """
        Setup a station data plot figure.
        """
        self.fig, axes_temp = plt.subplots(2, 2, sharex=True)
        self.axes = axes_temp.ravel().tolist()
        self.fig.set_size_inches((13.5, 7.0))

        # Have to deal with axes
        # Set log

        for ax in self.axes:
            ax.set_xscale("log")

        self.axes[0].invert_xaxis()
        self.axes[0].set_yscale("log")
        self.axes[2].set_yscale("log")
        # Set labels
        self.axes[2].set_xlabel("Frequency [Hz]")
        self.axes[3].set_xlabel("Frequency [Hz]")
        self.axes[0].set_ylabel("Apperent resistivity [Ohm m]")
        self.axes[1].set_ylabel("Apperent phase [degrees]")
        self.axes[1].set_ylim(-180, 180)
        self.axes[2].set_ylabel("Impedance amplitude [V/A]")
        self.axes[3].set_ylim(-180, 180)
        self.axes[3].set_ylabel("Impedance angle [degrees]")

    def draw(self, data_list, location):
        """
        Function to draw on the axes

        :param data_list: List of NSEM data objects to plot.
            Has to be of length >= 1. First item is treat as a
            observed data (Hast to have relative_error and noise_floor)
            assigned) and the others are plotted on top.
        :param location: Location of the station to plot
        """

        axes = self.axes

        # Pop the data from the list
        data = data_list[0]

        # Apparent resistivity
        data.plot_app_res(location, ["xy", "yx"], ax=axes[0], errorbars=True)

        # Apparent phase
        data.plot_app_phs(location, ["xy", "yx"], ax=axes[1], errorbars=True)

        # Impedamce amplitude
        data.plot_imp_amp(
            location, ["xx", "xy", "yx", "yy"], ax=axes[2], errorbars=True
        )

        # Impedance phase
        data.plot_app_phs(
            location, ["xx", "xy", "yx", "yy"], ax=axes[3], errorbars=True
        )

        # Plot the additional data
        for other_data in data_list[1::]:
            # Need add symbol generation
            dd_kwargs = {
                "xx": {"marker": ".", "ls": "--"},
                "xy": {"marker": ".", "ls": "--"},
                "yx": {"marker": ".", "ls": "--"},
                "yy": {"marker": ".", "ls": "--"},
            }

            # Apparent resistivity
            other_data.plot_app_res(
                location,
                ["xy", "yx"],
                ax=axes[0],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )

            # Apparent phase
            other_data.plot_app_phs(
                location,
                ["xy", "yx"],
                ax=axes[1],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )

            # Impedamce amplitude
            other_data.plot_imp_amp(
                location,
                ["xx", "xy", "yx", "yy"],
                ax=axes[2],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )

            # Impedance phase
            other_data.plot_app_phs(
                location,
                ["xx", "xy", "yx", "yy"],
                ax=axes[3],
                errorbars=False,
                comp_plot_dict=dd_kwargs,
            )


class DataNSEMPlotMethods(object):
    """
    Class container for properties and methods for
    plotting of NSEM data.

    """

    def __init__(self):
        """
        Just for reference
        """
        pass

    def plot_app_res(
        self,
        location,
        components=("xy", "yx"),
        ax=None,
        errorbars=False,
        comp_plot_dict=DEFAULT_COMP_DICT,
    ):
        """
        Plot apperent resistivity curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes.Axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes.Axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: bool
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict

        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("Apperent resistivity [Ohm m]")
        else:
            ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(location, comp, "app_res", ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp]
                )

                self.station_errorbars(location, comp, "app_res", ax=ax, **eb_kwargs)

        return ax

    def plot_app_phs(
        self,
        location,
        components=("xy", "yx"),
        ax=None,
        errorbars=False,
        comp_plot_dict=DEFAULT_COMP_DICT,
    ):
        """
        Plot apperent resistivity curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes.Axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes.Axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: bool
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot settings
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale("log")
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("Phase angle [Degrees]")
        else:
            ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(location, comp, "phase", ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp]
                )

                self.station_errorbars(location, comp, "phase", ax=ax, **eb_kwargs)

        return ax

    def plot_imp_amp(
        self,
        location,
        components=("xy", "yx"),
        ax=None,
        errorbars=False,
        comp_plot_dict=DEFAULT_COMP_DICT,
    ):
        """
        Plot impedance amplitude curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes.Axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes.Axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: bool
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("Impedance amplitude [V/A]")
        else:
            ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(location, comp, "amplitude", ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp]
                )

                self.station_errorbars(location, comp, "amplitude", ax=ax, **eb_kwargs)

        return ax

    def plot_tip_amp(
        self,
        location,
        components=("zx", "zy"),
        ax=None,
        errorbars=False,
        comp_plot_dict=DEFAULT_COMP_DICT,
    ):
        """
        Plot tipper amplitude curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes.Axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes.Axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: bool
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("Tipper magnitude [unitless]")
        else:
            ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(location, comp, "amplitude", ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp]
                )

                self.station_errorbars(location, comp, "amplitude", ax=ax, **eb_kwargs)

        return ax

    def map_data_locations(self, ax=None, **plot_kwargs):
        """
        Function that plots all receiver locations of the data
            (all discreate data locations).

        :param ax: The ax object for mapping to. Default: None
        :type ax: :class:`axes <matplotlib.axes.Axes>`

        """
        # Default plot dict
        default_dict = {"marker": "+", "c": "k", "ms": 10, "ls": "None", "zorder": 4}

        # Set the default plot kwargs
        for key, val in default_dict.items():
            # Make sure they aren't already assigned
            if key not in plot_kwargs:
                plot_kwargs[key] = val
        # Get unique locations
        unique_locations = _unique_rows(
            np.concatenate(
                [
                    rx.locations
                    for src in self.survey.source_list
                    for rx in src.receiver_list
                ]
            )
        )
        # Make the figure and the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the locations
        ax.plot(unique_locations[:, 0], unique_locations[:, 1], **plot_kwargs)

        return (fig, ax)

    def station_component(
        self, location, orientation, component, ax=None, **plot_kwargs
    ):
        """
        :param numpy.ndarray location: Coordnaties of the station to plot
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes.Axes ax: Axes (optional):
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs:

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        freqs, plot_data = _get_plot_data(self, location, orientation, component)
        # Plot
        plot_obj = ax.plot(freqs, plot_data, **plot_kwargs)

        return (fig, ax, plot_obj)

    def station_errorbars(
        self, location, orientation, component, ax=None, **plot_kwargs
    ):
        """
        :param numpy.ndarray location: Coordnaties of the station to plot
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes.Axes ax: Axes (optional)
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs:

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        freqs, plot_data, errorbars = _get_station_data(
            self, location, orientation, component, plot_error=True
        )
        plot_obj = ax.errorbar(freqs, plot_data, yerr=errorbars, **plot_kwargs)
        return (fig, ax, plot_obj)

    def frequency_map(
        self, frequency, orientation, component, ax=None, plot_error=True, **plot_kwargs
    ):
        """
        Function to generate a iso-frequency map

        :param numpy.ndarray frequency: Frequency to be mapped
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes.Axes ax: Axes (optional)
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs:

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        locs, plot_data, errorbars = _get_map_data(
            self, frequency, orientation, component, plot_error
        )
        plot_obj = plt.tricontourf(locs, plot_data, **plot_kwargs)
        return (fig, ax, plot_obj)


# Hidden utils functions
def _get_map_data(data, frequency, orientation, component, plot_error=False):
    """
    Function for getting frequency map data
    """

    # Get the components
    if component in ["app_res", "phase", "amplitude"]:
        real_tuple = _extract_frequency_data(
            data, frequency, orientation, "real", plot_error
        )
        imag_tuple = _extract_frequency_data(
            data, frequency, orientation, "imag", plot_error
        )
        if plot_error:
            freqs, real_data, real_std, real_floor = real_tuple
            freqs, imag_data, imag_std, imag_floor = imag_tuple
            # Add up the uncertainties
            real_uncert = real_std * np.abs(real_data) + real_floor
            imag_uncert = imag_std * np.abs(imag_data) + imag_floor
        else:
            freqs, real_data = real_tuple
            freqs, imag_data = imag_tuple

        if "app_res" in component:
            comp_data = real_data + 1j * imag_data
            plot_data = (1.0 / (mu_0 * omega(freqs))) * np.abs(comp_data) ** 2
            if plot_error:
                res_uncert = (2.0 / (mu_0 * omega(freqs))) * np.sqrt(
                    (real_data * real_uncert) ** 2 + (imag_data * imag_uncert) ** 2
                )
                errorbars = [res_uncert, res_uncert]
        elif "phase" in component:
            plot_data = np.arctan2(imag_data, real_data) * (180.0 / np.pi)
            if plot_error:
                phs_uncert = np.abs(
                    (1.0 / (real_data**2 + imag_data**2))
                    * ((real_data * real_uncert - imag_data * imag_uncert))
                ) * (180.0 / np.pi)
                # Scale back the errorbars
                errorbars = [phs_uncert, phs_uncert]
        elif "amplitude" in component:
            comp_data = real_data + 1j * imag_data
            plot_data = np.abs(comp_data)
            if plot_error:
                amp_uncert = (1.0 / plot_data) * (
                    (np.abs(real_data) * real_uncert)
                    + (np.abs(imag_data) * imag_uncert)
                )
                errorbars = [amp_uncert, amp_uncert]  # [low_unsert, up_unsert]
    else:
        if plot_error:
            freqs, plot_data, std_data, floor_data = _extract_frequency_data(
                data, frequency, orientation, component, return_uncert=error
            )
            attr_uncert = std_data * np.abs(plot_data) + floor_data
            errorbars = [attr_uncert, attr_uncert]
        else:
            freqs, plot_data = _extract_frequency_data(
                data, frequency, orientation, component, return_uncert=False
            )
    return (freqs, plot_data, errorbars)


def _get_station_data(data, location, orientation, component, plot_error=False):
    # Get the components
    if component in ["app_res", "phase", "amplitude"]:
        real_tuple = _extract_location_data(
            data, location, orientation, "real", plot_error
        )
        imag_tuple = _extract_location_data(
            data, location, orientation, "imag", plot_error
        )

        if plot_error:
            freqs, real_data, real_std, real_floor = real_tuple
            freqs, imag_data, imag_std, imag_floor = imag_tuple
            # Add up the uncertainties
            real_uncert = real_std * np.abs(real_data) + real_floor
            imag_uncert = imag_std * np.abs(imag_data) + imag_floor
        else:
            freqs, real_data = real_tuple
            freqs, imag_data = imag_tuple

        if "app_res" in component:
            comp_data = real_data + 1j * imag_data
            plot_data = (1.0 / (mu_0 * omega(freqs))) * np.abs(comp_data) ** 2
            if plot_error:
                res_uncert = (2.0 / (mu_0 * omega(freqs))) * np.sqrt(
                    (real_data * real_uncert) ** 2 + (imag_data * imag_uncert) ** 2
                )
                errorbars = [res_uncert, res_uncert]
        elif "phase" in component:
            plot_data = np.arctan2(imag_data, real_data) * (180.0 / np.pi)
            if plot_error:
                phs_uncert = np.abs(
                    (1.0 / (real_data**2 + imag_data**2))
                    * ((real_data * real_uncert - imag_data * imag_uncert))
                ) * (180.0 / np.pi)
                # Scale back the errorbars
                errorbars = [phs_uncert, phs_uncert]
        elif "amplitude" in component:
            comp_data = real_data + 1j * imag_data
            plot_data = np.abs(comp_data)
            if plot_error:
                amp_uncert = (1.0 / plot_data) * (
                    (np.abs(real_data) * real_uncert)
                    + (np.abs(imag_data) * imag_uncert)
                )
                errorbars = [amp_uncert, amp_uncert]  # [low_unsert, up_unsert]
    else:
        if plot_error:
            freqs, plot_data, std_data, floor_data = _extract_location_data(
                data, location, orientation, component, return_uncert=plot_error
            )
            attr_uncert = std_data * np.abs(plot_data) + floor_data
            errorbars = [attr_uncert, attr_uncert]
        else:
            freqs, plot_data = _extract_location_data(
                data, location, orientation, component, return_uncert=plot_error
            )
    if plot_error:
        return (freqs, plot_data, errorbars)
    else:
        return (freqs, plot_data)


def _get_plot_data(data, location, orientation, component):
    if "app_res" in component:
        freqs, dat_r = _extract_location_data(data, location, orientation, "real")
        freqs, dat_i = _extract_location_data(data, location, orientation, "imag")
        dat = dat_r + 1j * dat_i
        plot_data = 1.0 / (mu_0 * omega(freqs)) * np.abs(dat) ** 2
    elif "phase" in component:
        freqs, dat_r = _extract_location_data(data, location, orientation, "real")
        freqs, dat_i = _extract_location_data(data, location, orientation, "imag")
        plot_data = np.arctan2(dat_i, dat_r) * (180.0 / np.pi)
    elif "amplitude" in component:
        freqs, dat_r = _extract_location_data(data, location, orientation, "real")
        freqs, dat_i = _extract_location_data(data, location, orientation, "imag")
        dat_complex = dat_r + 1j * dat_i
        plot_data = np.abs(dat_complex)
    else:
        freqs, plot_data = _extract_location_data(
            data, location, orientation, component
        )
    return (freqs, plot_data)


def _extract_frequency_data(
    data, frequency, orientation, component, return_uncert=False
):
    """
    Function to extract data at given frequency
    """

    src = data.survey.get_sources_by_frequency(frequency)
    rx_list = [
        rx
        for rx in src.receiver_list
        if rx.orientation == orientation and rx.component == component
    ]
    # Check the number of the rx
    if len(rx_list) == 1:
        rx = rx_list[0]
    elif len(rx_list) == 0:
        # Should add a warning that the rx doesn't excist
        if return_uncert:
            # Returning all empty arrays
            return (np.array([]), np.array([]), np.array([]), np.array([]))
        return (np.array([]), np.array([]))
    else:
        # Should be a more specifice Exeption
        raise Exception("To many Receivers of the same type, orientation and component")

    loc_arr = rx.locations
    data_arr = data[src, rx]
    if return_uncert:
        std_arr = data.relative_error[src, rx]
        floor_arr = data.floor[src, rx]
    if return_uncert:
        return (loc_arr, data_arr, std_arr, floor_arr)
    return (loc_arr, data_arr)


def _extract_location_data(data, location, orientation, component, return_uncert=False):
    """
    Function to extract data at given location
    """
    freq_list = []
    data_list = []
    std_list = []
    floor_list = []
    for src in data.survey.source_list:
        rx_list = [
            rx
            for rx in src.receiver_list
            if rx.orientation == orientation and rx.component == component
        ]
        if len(rx_list) == 0:
            if return_uncert:
                return (np.array([]), np.array([]), np.array([]), np.array([]))
            return (np.array([]), np.array([]))
        else:
            rx = rx_list[0]

        ind_loc = np.sqrt(np.sum((rx.locations[:, :2] - location) ** 2, axis=1)) < 0.1
        if np.any(ind_loc):
            freq_list.append(src.frequency)
            data_list.append(data[src, rx][ind_loc])

            if return_uncert:
                index = data.index_dictionary[src][rx]
                std_list.append(data.relative_error[index][ind_loc])
                floor_list.append(data.noise_floor[index][ind_loc])
    if return_uncert:
        return (
            np.array(freq_list),
            np.concatenate(data_list),
            np.concatenate(std_list),
            np.concatenate(floor_list),
        )
    return (np.array(freq_list), np.concatenate(data_list))


## Hidden utility functions
# Unique row function-should be moved to utils
def _unique_rows(array):
    """
    Finds and returns unique rows in an array
    """
    array = np.ascontiguousarray(array)
    unique_array = np.unique(array.view([("", array.dtype)] * array.shape[1]))
    return unique_array.view(array.dtype).reshape(
        (unique_array.shape[0], array.shape[1])
    )
